import io
import json
import logging
import socket
import threading
import zipfile
import time
from queue import Queue

import requests
from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes

from biolib.compute_node.socker_listener_thread import SocketListenerThread
from biolib.compute_node.socket_sender_thread import SocketSenderThread
from biolib.compute_node.compute_process.biolib_docker_client import BiolibDockerClient
from biolib.compute_node.compute_process.mappings import Mappings, path_without_first_folder
from biolib.compute_node.compute_process.utils import ComputeProcessException
from biolib.compute_node.utils import get_package_type, SystemExceptionCodes, enclave_remote_hosts
from biolib.biolib_binary_format import AttestationDocument, SavedJob, SystemStatusUpdate, ModuleInput, ModuleOutput,\
    RsaEncryptedAesPackage, AesEncryptedPackage

try:
    from biolib.compute_node.compute_process.nsm_util import NSMUtil
except ImportError:
    pass

DEFAULT_BUFFER_SIZE = 1024
SOCKET_HOST = '127.0.0.1'


class ComputeProcess:
    def __init__(self, socket_port, is_running_in_enclave):
        try:
            self.is_running_in_enclave = is_running_in_enclave
            self.biolib_logger = logging.getLogger('biolib')
            if self.is_running_in_enclave:
                self.socket_port = 5005
                self.nsm_util = NSMUtil()
                self.aes_key_buffer = b''
                self.biolib_logger.setLevel(logging.DEBUG)
            else:
                self.socket_port = socket_port
            self.received_messages_queue = Queue()
            self.messages_to_send_queue = Queue()
            self.biolib_docker_client = BiolibDockerClient(self.is_running_in_enclave, self.messages_to_send_queue)
            self.access_token = ''
            self.biolib_host = ''
            self.ecr_proxy = ''
            self.enclave_ecr_token = ''
            self.job_id = ''
            self.remote_hostnames = []
            self.runtime_zip_data = None
            self.module = None
            self.locks = {
                'runtime_zip': threading.Lock(),
                'remote_host_proxies': threading.Lock(),
                'image': threading.Lock()
            }

        except Exception as exception:
            raise ComputeProcessException(exception,
                                          SystemExceptionCodes.FAILED_TO_INIT_COMPUTE_PROCESS_VARIABLES.value,
                                          self.messages_to_send_queue) from exception

        self.biolib_docker_client.create_networks()
        if self.is_running_in_enclave:
            self.biolib_docker_client.start_remote_host_proxies(enclave_remote_hosts, self.locks['remote_host_proxies'])
            # Block until all proxies are created
            self.locks['remote_host_proxies'].acquire()
            self.locks['remote_host_proxies'].release()

        self.connect_to_parent()

    def run(self):
        self.biolib_logger.debug("Started Compute Process")
        while True:
            try:
                package = self.received_messages_queue.get()
                if package == b'STOP':
                    self.biolib_docker_client.delete_old_state()
                    break

                package_type = get_package_type(package)
                if package_type == 'RsaEncryptedAesPackage':
                    encrypted_aes_key, iv, _, encrypted_data = RsaEncryptedAesPackage(package).deserialize()
                    self.aes_key_buffer = self.nsm_util.decrypt(encrypted_aes_key)
                    aes_key = AES.new(self.aes_key_buffer, AES.MODE_GCM, iv)

                    package = aes_key.decrypt(encrypted_data)
                    package_type = get_package_type(package)

                if package_type == 'SavedJob':
                    runtime_zip_link = self.get_runtime_zip_link_and_set_fields_from_job_package(package)
                    if runtime_zip_link:
                        self.send_status_update(progress=25, log_message='Downloading Source Files...')
                        self.get_runtime_zip(runtime_zip_link)

                    if self.remote_hostnames:
                        self.send_status_update(progress=35,
                                                log_message='Starting Remote Hosts...')
                        self.biolib_docker_client.start_remote_host_proxies(self.remote_hostnames,
                                                                            self.locks['remote_host_proxies'])

                elif package_type == 'ModuleInput':
                    module_input = ModuleInput(package).deserialize()
                    self.start_compute(module_input, self.job_id)

                else:
                    self.biolib_logger.warning(f'Package type from parent was not recognized: {package}')

                self.received_messages_queue.task_done()
            except ComputeProcessException:
                continue

            except Exception as exception:
                raise ComputeProcessException(exception, SystemExceptionCodes.UNKOWN_COMPUTE_PROCESS_ERROR.value,
                                              self.messages_to_send_queue) from exception

    def start_compute(self, module_input, job_id):
        self.block_until_data_is_available_for_compute()
        if self.module['environment'] == 'biolib-app':
            job = self.get_job_for_external_app()
            job_id = job['public_id']
            module_input_with_runtime_code = self.add_runtime_zip_and_command_to_module_input(module_input)

            if 'client_side_executable_zip' in job['app_version']:
                self.get_runtime_zip(job['app_version']['client_side_executable_zip'])

            self.module = self.get_module_from_name(job['app_version']['modules'], 'main')
            self.start_compute(module_input_with_runtime_code, job_id)
            return

        self.send_status_update(progress=55, log_message='Pulling images...')
        self.biolib_docker_client.pull_image(self.module['image_uri'], self.access_token, job_id,
                                             self.locks['image'], self.ecr_proxy, self.enclave_ecr_token)
        threading.Thread(target=self._compute, args=(module_input,)).start()

    def _compute(self, module_input):
        # Block until the image has been pulled
        self.locks['image'].acquire()
        self.send_status_update(progress=70, log_message='Computing...')
        start_time = time.time()

        self.biolib_docker_client.create_compute_container(
            image=f'{self.ecr_proxy}/{self.module["image_uri"]}',
            command=f'{self.module["command"]} {" ".join(module_input["arguments"])}',
            working_dir=self.module['working_directory']
        )

        # Also pass arguments so we can parse $ variables when mapping later
        self.biolib_docker_client.set_mappings(
            self.module['input_files_mappings'],
            self.module['source_files_mappings'],
            self.module['output_files_mappings'],
            module_input['arguments']
        )

        if self.runtime_zip_data:
            self.biolib_docker_client.map_and_copy_runtime_files_to_container(self.runtime_zip_data)

        self.biolib_docker_client.map_and_copy_input_files_to_container(module_input['files'])

        stdout, stderr, exit_code, mapped_output_files = self.biolib_docker_client.run_container()

        try:
            module_output = ModuleOutput().serialize(stdout, stderr, exit_code, mapped_output_files)
            if self.is_running_in_enclave:
                module_output_to_send = self.wrap_in_aes_encrypted_package(module_output)
            else:
                module_output_to_send = module_output

            self.biolib_logger.debug(f'Compute time: {time.time() - start_time}')
            self.messages_to_send_queue.put(module_output_to_send)

        except Exception as exception:
            raise ComputeProcessException(exception,
                                          SystemExceptionCodes.FAILED_TO_SERIALIZE_AND_SEND_MODULE_OUTPUT.value,
                                          self.messages_to_send_queue) from exception

    def connect_to_parent(self):
        try:
            if self.is_running_in_enclave:
                parent_socket = socket.socket(socket.AF_VSOCK, socket.SOCK_STREAM)  # pylint: disable=no-member
                parent_socket.connect((3, int(self.socket_port)))

            else:
                parent_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                parent_socket.connect((SOCKET_HOST, int(self.socket_port)))

        except Exception as exception:
            raise ComputeProcessException(exception,
                                          SystemExceptionCodes.FAILED_TO_CONNECT_TO_WORKER_THREAD_SOCKET.value,
                                          self.messages_to_send_queue) from exception

        try:
            SocketListenerThread(parent_socket, self.received_messages_queue).start()
            SocketSenderThread(parent_socket, self.messages_to_send_queue).start()
        except Exception as exception:
            raise ComputeProcessException(exception,
                                          SystemExceptionCodes.FAILED_TO_START_SENDER_THREAD_OR_RECEIVER_THREAD.value,
                                          self.messages_to_send_queue) from exception

        try:
            if self.is_running_in_enclave:
                attestation_document = self.nsm_util.get_attestation_doc()
            else:
                attestation_document = b'Running locally'
        except Exception as exception:
            raise ComputeProcessException(exception, SystemExceptionCodes.FAILED_TO_GET_ATTESTATION_DOCUMENT.value,
                                          self.messages_to_send_queue) from exception

        self.messages_to_send_queue.put(AttestationDocument().serialize(attestation_document))

    def get_runtime_zip_link_and_set_fields_from_job_package(self, package):
        try:
            saved_job_json_string = SavedJob(package).deserialize()
            saved_job = json.loads(saved_job_json_string)

            remote_hosts = saved_job['job']['app_version']['remote_hosts']
            self.remote_hostnames = [remote_host['hostname'] for remote_host in remote_hosts]
            self.job_id = saved_job['job']['public_id']
            self.biolib_host = saved_job['BIOLIB_HOST']
            if self.biolib_host == 'https://biolib.com':
                self.ecr_proxy = 'containers.biolib.com'
            else:
                self.ecr_proxy = 'containers.staging.biolib.com'

            if 'enclave_ecr_token' in saved_job:
                self.enclave_ecr_token = saved_job['enclave_ecr_token']

            if 'access_token' in saved_job:
                self.access_token = saved_job['access_token']

            module_name = saved_job['module_name']
            self.module = self.get_module_from_name(saved_job['job']['app_version']['modules'], module_name)

            if 'client_side_executable_zip' in saved_job['job']['app_version']:
                return saved_job['job']['app_version']['client_side_executable_zip']
            else:
                return ''

        except Exception as exception:
            raise ComputeProcessException(exception, SystemExceptionCodes.FAILED_TO_DESERIALIZE_SAVED_JOB.value,
                                          self.messages_to_send_queue) from exception

    def get_job_for_external_app(self):
        try:
            data = {'app_version_id': self.module['image_uri']}
            if self.access_token:
                headers = {'Authentication': f'Bearer {self.access_token}'}
            else:
                headers = {}
            req = requests.post(f'{self.biolib_host}/api/jobs/', json=data, headers=headers)
        except Exception as exception:
            raise ComputeProcessException(exception, SystemExceptionCodes.FAILED_TO_CONTACT_BACKEND_TO_CREATE_JOB.value,
                                          self.messages_to_send_queue) from exception

        if req.status_code != 201:
            raise ComputeProcessException(Exception(req.content), SystemExceptionCodes.FAILED_TO_CREATE_NEW_JOB.value,
                                          self.messages_to_send_queue)
        return req.json()

    def add_runtime_zip_and_command_to_module_input(self, module_input):
        # TODO: Figure out if we ever forward output mappings correctly (Do we only the mapping of the base image?)
        # TODO: Reuse much of the make_runtime_tar logic in BiolibDockerClient
        try:
            if self.runtime_zip_data:
                runtime_zip = zipfile.ZipFile(io.BytesIO(self.runtime_zip_data))
                source_mappings = Mappings(self.module['source_files_mappings'], module_input['arguments'])
                for zip_file_name in runtime_zip.namelist():
                    file_path = '/' + path_without_first_folder(zip_file_name)
                    mapped_file_names = source_mappings.get_mappings_for_path(file_path)
                    for mapped_file_name in mapped_file_names:
                        file_data = runtime_zip.read(zip_file_name)
                        module_input['files'].update({mapped_file_name: file_data})

            for command_part in reversed(self.module['command'].split()):
                module_input['arguments'].insert(0, command_part)

        except Exception as exception:
            raise ComputeProcessException(exception, SystemExceptionCodes.FAILED_TO_CREATE_NEW_JOB.value,
                                          self.messages_to_send_queue) from exception
        return module_input

    def get_runtime_zip(self, runtime_zip_link):
        # TODO: Implement caching of runtime zip
        try:
            self.locks['runtime_zip'].acquire()
            threading.Thread(target=self._download_runtime_zip_from_s3, args=[runtime_zip_link]).start()
        except Exception as exception:
            raise ComputeProcessException(exception,
                                          SystemExceptionCodes.FAILED_TO_START_RUNTIME_ZIP_DOWNLOAD_THREAD.value,
                                          self.messages_to_send_queue) from exception

    def _download_runtime_zip_from_s3(self, runtime_zip_link):
        start_time = time.time()
        self.biolib_logger.debug(f"Downloading runtime zip from s3 link: {runtime_zip_link}")
        try:
            self.runtime_zip_data = requests.get(runtime_zip_link).content
        except Exception as exception:
            raise ComputeProcessException(exception, SystemExceptionCodes.FAILED_TO_DOWNLOAD_RUNTIME_ZIP.value,
                                          self.messages_to_send_queue) from exception
        self.biolib_logger.debug(f'Downloading runtime zip took: {time.time() - start_time}s')
        self.locks['runtime_zip'].release()

    def block_until_data_is_available_for_compute(self):
        # Block until the required data is available
        try:
            for lock_name, lock in self.locks.items():
                self.biolib_logger.debug(f"Waiting for {lock_name}")
                lock.acquire()
                lock.release()
        except Exception as exception:
            raise ComputeProcessException(exception, SystemExceptionCodes.FAILED_TO_GET_REQUIRED_DATA_FOR_COMPUTE.value,
                                          self.messages_to_send_queue) from exception

    def get_module_from_name(self, modules, module_name):
        for module in modules:
            if module['name'] == module_name:
                return module
        raise Exception(f"Could not find module with name {module_name} in given modules")

    def wrap_in_aes_encrypted_package(self, package):
        iv = get_random_bytes(12)
        aes_key = AES.new(self.aes_key_buffer, AES.MODE_GCM, iv)
        encrypted_package, tag = aes_key.encrypt_and_digest(package)
        aes_encrypted_package = AesEncryptedPackage().serialize(iv, tag, encrypted_package)
        return aes_encrypted_package

    def send_status_update(self, progress, log_message):
        try:
            status_update_package = SystemStatusUpdate().serialize(progress, log_message)
            self.biolib_logger.info(log_message)
            self.messages_to_send_queue.put(status_update_package)
        except Exception as exception:
            raise ComputeProcessException(exception, SystemExceptionCodes.FAILED_TO_SEND_STATUS_UPDATE.value,
                                          self.messages_to_send_queue) from exception



import re

class Mappings:
    def __init__(self, mappings_list, arguments):
        self.mappings_dict = {}
        for mapping in mappings_list:
            from_path, to_path = self.replace_dollar_variables_and_validate(mapping['from_path'],
                                                                            mapping['to_path'], arguments)
            self.mappings_dict[from_path] = to_path

    def get_mappings_for_path(self, path):
        mapped_file_names = []
        for from_path, to_path in self.mappings_dict.items():
            if (from_path.endswith('/') and path.startswith(from_path)) or path == from_path:

                # Mapping from dir to dir
                if from_path.endswith('/'):
                    mapped_file_names.append(to_path + path[len(from_path):])

                # Mapping from a file
                else:
                    # Mapping file to dir
                    if to_path.endswith('/'):
                        mapped_file_names.append(to_path + path.split('/')[-1])
                    # Mapping file to file
                    else:
                        mapped_file_names.append(to_path)

        return mapped_file_names

    def replace_dollar_variables_and_validate(self, from_path, to_path, arguments):
        from_path = self.replace_dollar_with_arg(from_path, arguments)
        to_path = self.replace_dollar_with_arg(to_path, arguments)

        assert_path_validity(from_path)
        assert_path_validity(to_path)

        return from_path, to_path

    def replace_dollar_with_arg(self, path, args):
        for arg_index in re.findall(r'\$(\d+)', path):
            if arg_index == '0':
                raise Exception('Attempted to substitute argument 0 which is invalid')

            if arg_index < len(args):
                raise Exception('Referred to an argument index out of bounds of arguments')

            path = path.replace(f'${arg_index}', args[int(arg_index) - 1])
        return path


def path_without_first_folder(path):
    """
    Removes the first folder in a relative path.
    example/test.txt -> test.txt
    some_dir/some_sub_dir/test.txt -> some_sub_dir/test.txt
    """
    return path.split('/', 1)[1]


def assert_path_validity(path: str):
    if not path.startswith('/'):
        raise Exception('Path must be absolute')

    if '//' in path:
        raise Exception('Path can not include consecutive slashes')

    if '/..' in path:
        raise Exception('Path can not include ".."')



# Importing ascii is necessary for file open to work in some environments
from encodings import ascii  # pylint: disable=redefined-builtin, unused-import
from pathlib import Path
import http.server
import asyncio
import base64
import logging
import os
import re
import shutil
import signal
import socketserver
import subprocess
import sys
import threading

import docker
import nest_asyncio  # necessary import fix required for async to work in notebooks
import requests
import yaml

from biolib.biolib_binary_format import ModuleInput, ModuleOutput
from biolib.pyppeteer.pyppeteer import launch, command
from biolib.pyppeteer.pyppeteer.launcher import resolveExecutablePath, __chromium_revision__
from biolib.compute_node.webserver import webserver
from biolib.compute_node.compute_process.compute_process import ComputeProcess

nest_asyncio.apply()

logging.basicConfig(format='%(asctime)s | %(levelname)s : %(message)s',
                    level=logging.INFO, stream=sys.stdout)
logging.getLogger('biolib.pyppeteer.pyppeteer').setLevel(logging.ERROR)
logging.getLogger('biolib.pyppeteer.pyppeteer.connection').setLevel(logging.ERROR)

biolib_logger = logging.getLogger('biolib')


class BioLibError(Exception):
    # pylint: disable=super-init-not-called
    def __init__(self, message):
        self.message = message


class CompletedProcess:
    def __init__(self, stdout, stderr, exitcode):
        self.stdout = stdout
        self.stderr = stderr
        self.exitcode = exitcode

    def ipython_markdown(self):
        from IPython.display import display, Markdown  # pylint: disable=import-error, import-outside-toplevel
        markdown_str = self.stdout.decode('utf-8')
        # prepend ./biolib_results/ to all paths
        # ie [SeqLogo](./SeqLogo2.png) test ![SeqLogo](./SeqLogo.png)
        # ![SeqLogo](SeqLogo.png)  ![SeqLogo](/SeqLogo.png)
        # is transformed to ![SeqLogo](./biolib_results/SeqLogo2.png) test ![SeqLogo](./biolib_results/SeqLogo.png)
        # ![SeqLogo](./biolib_results/SeqLogo.png)  ![SeqLogo](./biolib_results/SeqLogo.png)
        markdown_str_modified = re.sub(r'\!\[([^\]]*)\]\((\.\/|\/|)([^\)]*)\)',
                                       r'![\\1](./biolib_results/\\3)',
                                       markdown_str)
        display(Markdown(markdown_str_modified))


class BioLibServer(socketserver.TCPServer):
    log_level = logging.INFO


class BioLibHandler(http.server.SimpleHTTPRequestHandler):
    def log_message(self, format, *args):  # pylint: disable=redefined-builtin
        if self.server.log_level == BioLib.TRACE_LOGGING:
            http.server.SimpleHTTPRequestHandler.log_message(self, format, *args)


def js_to_python_byte_array_converter(js_encoded):
    try:
        return bytes(list([js_encoded[str(i)] for i in range(len(js_encoded))]))
    except Exception as error:
        biolib_logger.error("Failed to decode response from browser")
        biolib_logger.error(error)
        biolib_logger.error(js_encoded)
        raise BioLibError(js_encoded) from error


def python_bytes_to_byte64(data):
    return base64.b64encode(data).decode('ascii')


class BioLib:
    executable_path = None
    no_sandbox = True
    log_level = logging.INFO
    TRACE_LOGGING = 50
    auth_api_token = None
    auth_access_token = None
    auth_refresh_token = None
    host = 'https://biolib.com'

    @staticmethod
    def set_chrome_path(path):
        BioLib.executable_path = path

    @staticmethod
    def set_sandbox(use_sandbox):
        BioLib.no_sandbox = not use_sandbox

    @staticmethod
    def set_api_token(api_token):
        BioLib.auth_api_token = api_token

    @staticmethod
    def attempt_login(exit_on_failure=False):
        if BioLib.auth_api_token is None:
            if exit_on_failure:
                biolib_logger.error('Error: Attempted login, but BIOLIB_TOKEN was not set, exiting...')
                sys.exit(1)
            else:
                biolib_logger.debug('Attempted login, but BIOLIB_TOKEN was not set, so continuing without logging in')
                return

        response = requests.post(f'{BioLib.host}/api/user/api_tokens/exchange/',
                                 json={'token': BioLib.auth_api_token})
        json_response = response.json()
        if not response.ok:
            biolib_logger.error('Login with API token failed:')
            biolib_logger.error(json_response['detail'])
            sys.exit(1)
        else:
            BioLib.auth_refresh_token = json_response['refresh_token']
            BioLib.auth_access_token = json_response['access_token']
            biolib_logger.info('Successfully authenticated')

    @staticmethod
    def push(author_and_app_name, app_path):
        # prepare zip file
        if not Path(f'{app_path}/.biolib/config.yml').is_file():
            biolib_logger.error('The file .biolib/config.yml was not found in the application directory')
            sys.exit(1)
        cwd = os.getcwd()
        os.chdir(app_path)
        zip_name = 'biolib-cli-build-tmp-zip-1js7ah2akd'
        zip_filename = f'{zip_name}.zip'
        if Path(zip_filename).is_file():
            biolib_logger.error(
                f'Please make sure you do not have a file called {zip_filename} in the application directory')
            sys.exit(1)
        app_folder_name = Path(app_path).absolute().parts[-1]
        shutil.make_archive(zip_name, 'zip', root_dir='..', base_dir=app_folder_name)
        zip_file = open(zip_filename, 'rb')
        zip_binary = zip_file.read()
        zip_file.close()
        os.remove(zip_filename)
        # change back to old working directory
        os.chdir(cwd)

        # login and get app data
        BioLib.attempt_login(exit_on_failure=True)
        author_and_app_name_split = author_and_app_name.split('/')
        app = BioLib.get_app_from_author_and_app_name(author_and_app_name_split[0],
                                                      author_and_app_name_split[1])
        # push new app version
        response = requests.post(
            f'{BioLib.host}/api/app_versions/',
            headers={'Authorization': f'Bearer {BioLib.auth_access_token}'},
            files={
                'app': (None, app['public_id']),
                'set_as_active': (None, 'true'),
                'state': (None, 'published'),
                'source_files_zip': zip_binary
            })
        if not response.ok:
            biolib_logger.error(f'Push failed for {author_and_app_name}:')
            biolib_logger.error(response.text)
            sys.exit(1)
        else:
            # TODO: When response includes the version number, print the URL for the new app version
            json_response = response.json()
            biolib_logger.info(f'Successfully pushed app version for {author_and_app_name}.')

        docker_tags = json_response.get('docker_tags', {})
        if docker_tags:
            biolib_logger.info('Found docker images to push.')

            try:
                yaml_file = open(f'{app_path}/.biolib/config.yml', 'r', encoding='utf-8')

            except Exception as error:  # pylint: disable=broad-except
                biolib_logger.debug(error)
                biolib_logger.error('Could not open the config file .biolib/config.yml')
                sys.exit(1)

            try:
                config_data = yaml.safe_load(yaml_file)

            except Exception as error:  # pylint: disable=broad-except
                biolib_logger.debug(error)
                biolib_logger.error('Could not parse .biolib/config.yml. Please make sure it is valid YAML')
                sys.exit(1)

            # Auth to be sent to proxy
            # The tokens are sent as "{access_token},{job_id}". We leave job_id blank on push.
            tokens = f'{BioLib.auth_access_token},'
            auth_config = {'username': 'biolib', 'password': tokens}

            # the final step of docker push can take a long time, so set a long timeout for docker
            # 15 min (900s) is the maximum supported by AWS load balancers
            docker_client = docker.from_env(timeout=900)

            for module_name, repo_and_tag in docker_tags.items():
                docker_image_definition = config_data['modules'][module_name]['image']
                repo, tag = repo_and_tag.split(':')

                if docker_image_definition.startswith('dockerhub://'):
                    docker_image_name = docker_image_definition.replace('dockerhub://', 'docker.io/', 1)
                    biolib_logger.info(
                        f'Pulling image {docker_image_name} defined on module {module_name} from Dockerhub.')
                    dockerhub_repo, dockerhub_tag = docker_image_name.split(':')
                    docker_client.images.pull(dockerhub_repo, tag=dockerhub_tag)

                elif docker_image_definition.startswith('local-docker://'):
                    docker_image_name = docker_image_definition.replace('local-docker://', '', 1)

                try:
                    biolib_logger.info(f'Trying to push image {docker_image_name} defined on module {module_name}.')
                    image = docker_client.images.get(docker_image_name)
                    ecr_proxy_host = f'containers.{BioLib.host.split("://")[1]}/{repo}'
                    image.tag(ecr_proxy_host, tag)
                    for line in docker_client.images.push(ecr_proxy_host, tag=tag, stream=True,
                                                          decode=True, auth_config=auth_config):
                        biolib_logger.info(line)

                except Exception as exception:  # pylint: disable=broad-except
                    biolib_logger.debug(exception)
                    biolib_logger.error(f'Failed to tag and push image {docker_image_name}.')
                    sys.exit(1)

                biolib_logger.info(f'Successfully pushed {docker_image_name}')

            biolib_logger.info('Successfully pushed all docker images')

    @staticmethod
    def build_enclave(dev_mode):
        compute_node_dir = f'{BioLib.get_pybiolib_root()}/biolib/compute_node'
        output_eif_path = f'{os.getcwd()}/biolib-nitro-enclave.eif'

        if os.path.exists(output_eif_path):
            os.remove(output_eif_path)

        if dev_mode:
            pybiolib_tar_path = f'{BioLib.get_pybiolib_root()}/biolib/compute_node/enclave_build/pybiolib-0.0.1.tar.gz'
            if os.path.exists(pybiolib_tar_path):
                os.remove(pybiolib_tar_path)

            cur_dir = os.getcwd()
            os.chdir(BioLib.get_pybiolib_root())
            subprocess.run(['poetry', 'build'], check=True)
            shutil.move('./dist/pybiolib-0.0.1.tar.gz', pybiolib_tar_path)
            os.chdir(cur_dir)

            subprocess.run(
                ['sudo', 'docker', 'build', f'{compute_node_dir}/enclave_build/', '-t', 'biolib-nitro-enclave',
                 '--build-arg', 'ENV=dev'],
                check=True)
        else:
            subprocess.run(
                ['sudo', 'docker', 'build', f'{compute_node_dir}/enclave_build/', '-t', 'biolib-nitro-enclave'],
                check=True)

        subprocess.run(
            ['sudo', 'nitro-cli', 'build-enclave', '--docker-uri', 'biolib-nitro-enclave:latest', '--output-file',
             output_eif_path],
            check=True)

    @staticmethod
    def start_compute_node(port, host, eif_path=None):
        webserver.start_webserver(port=port, host=host, specified_biolib_host=BioLib.host, specified_eif_path=eif_path)

    @staticmethod
    def start_compute_process(socket_port, is_running_in_enclave):
        ComputeProcess(socket_port, is_running_in_enclave).run()
        # Kill the compute process which is the current process
        os.kill(os.getpid(), signal.SIGKILL)

    @staticmethod
    def get_app_from_author_and_app_name(author_name, app_name):
        headers = {}
        if BioLib.auth_access_token is not None:
            headers['Authorization'] = f'Bearer {BioLib.auth_access_token}'
        apps_response = requests.get(f'{BioLib.host}/api/apps/?account_handle={author_name}&app_name={app_name}',
                                     headers=headers)
        results = apps_response.json()['results']
        if len(results) == 0:
            raise Exception(f'Application "{author_name}/{app_name}" was not found. '
                            f'Make sure you have spelled the application name correctly, '
                            f'and that you have the required permissions.')
        return results[0]

    @staticmethod
    def get_pybiolib_root():
        # Use dirname() twice to get pybiolib/ as __file__ pybiolib/biolib/app.py
        return os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

    @staticmethod
    def set_logging(log_level):
        BioLib.log_level = log_level
        # only activate debug logging globally if user selected trace logging
        if log_level == BioLib.TRACE_LOGGING:
            logging.getLogger().setLevel(logging.DEBUG)
        elif log_level == logging.DEBUG:
            logging.getLogger().setLevel(logging.INFO)
            biolib_logger.setLevel(logging.DEBUG)
        else:
            logging.getLogger().setLevel(log_level)

    @staticmethod
    def set_host(host):
        if host:
            BioLib.host = host


# set default logging to info
BioLib.set_logging(logging.INFO)


class BioLibApp:
    app_author = None
    app_name = None

    def __init__(self, author, name):
        self.app_author = author
        self.app_name = name
        self.app_version_public_id = None

    async def call_pyppeteer(self, port, args, stdin, files, output_dir_path):
        if not BioLib.executable_path:
            mac_chrome_path = '/Applications/Google Chrome.app/Contents/MacOS/Google Chrome'
            if os.path.isfile(mac_chrome_path):
                BioLib.set_chrome_path(mac_chrome_path)

        if not BioLib.executable_path:
            linux_chrome_path = '/usr/lib/chromium-browser/chromium-browser'

            # special install for google colab
            if not os.path.isfile(linux_chrome_path) and 'google.colab' in sys.modules:
                subprocess.run('apt-get update', shell=True, check=True)
                subprocess.run('apt install chromium-chromedriver', shell=True, check=True)

            if os.path.isfile(linux_chrome_path):
                BioLib.set_chrome_path(linux_chrome_path)

        resolved_path = resolveExecutablePath(None, __chromium_revision__)
        if not BioLib.executable_path and resolved_path[1]:
            # if executable_path is not set explicit,
            # and resolveExecutablePath failed (== we got an error message back in resolved_path[1])
            logging.info('Installing dependencies...')
            os.environ['PYPPETEER_NO_PROGRESS_BAR'] = 'true'
            command.install()

        logging.info('Computing...')

        chrome_arguments = [
            '--disable-web-security',
        ]
        if BioLib.no_sandbox:
            chrome_arguments.append('--no-sandbox')

        browser = await launch(args=chrome_arguments, executablePath=BioLib.executable_path)

        # start new page
        page = await browser.newPage()

        await page.goto('http://localhost:' + str(port))

        def get_data():
            input_serialized = ModuleInput().serialize(stdin, args, files)
            return python_bytes_to_byte64(input_serialized)

        def set_progress_compute(value):
            biolib_logger.debug(f'Compute progress: {value}')

        def set_progress_initialization(value):
            biolib_logger.debug(f'Initialization progress: {value}')

        def add_log_message(value):
            biolib_logger.debug(f'Log message: {value}')

        await page.exposeFunction('getData', get_data)
        await page.exposeFunction('setProgressCompute', set_progress_compute)
        await page.exposeFunction('setProgressInitialization', set_progress_initialization)
        await page.exposeFunction('addLogMessage', add_log_message)

        refresh_token = 'undefined'
        if BioLib.auth_refresh_token:
            refresh_token = f'\'{BioLib.auth_refresh_token}\''

        output_serialized_js_bytes = await page.evaluate('''
        async function() {
          const refreshToken = ''' + refresh_token + ''';
          const appVersionId = \'''' + self.app_version_public_id + '''\';
          const baseUrl = \'''' + BioLib.host + '''\';

          const { BioLibSingleton, AppClient } = window.BioLib;
          BioLibSingleton.setConfig({ baseUrl, refreshToken });
          AppClient.setApiClient(BioLibSingleton.get());

          const inputBase64 = await window.getData();
          const inputByteArray = Uint8Array.from(atob(inputBase64), c => c.charCodeAt(0));

          const jobUtils = {
              setProgressCompute: window.setProgressCompute,
              setProgressInitialization: window.setProgressInitialization,
              addLogMessage: window.addLogMessage,
          };

          try {
            const moduleOutput = await AppClient.runAppVersion(appVersionId, inputByteArray, jobUtils);
            return moduleOutput.serialize();
          } catch(err) {
            return err.toString();
          }
        }
        ''')
        logging.debug('Closing browser')
        await browser.close()

        output_serialized = js_to_python_byte_array_converter(output_serialized_js_bytes)
        output = ModuleOutput(output_serialized).deserialize()

        if isinstance(output, dict):
            if len(output['files']) > 0:
                logging.info(f'Writing output files to: {output_dir_path}')

                for file_path, data in output['files'].items():
                    full_path = output_dir_path + file_path

                    if full_path.endswith('/'):
                        path_without_file_name = full_path
                        filename = ''
                    else:
                        full_path_parts = full_path.split('/')
                        filename = full_path_parts.pop()
                        path_without_file_name = '/'.join(full_path_parts) + '/'

                    os.makedirs(path_without_file_name, exist_ok=True)

                    if filename != '':
                        output_file = open(path_without_file_name + filename, 'wb')
                        output_file.write(data)
                        output_file.close()

            return CompletedProcess(stdout=output['stdout'], stderr=output['stderr'], exitcode=output['exit_code'])
        else:
            raise BioLibError(output)

    def __call__(self, args=None, stdin=None, files=None, output_path='biolib_results'):
        if args is None:
            args = []
        BioLib.attempt_login()
        logging.info('Loading package...')
        app = BioLib.get_app_from_author_and_app_name(self.app_author, self.app_name)
        active_version = app['active_version']
        self.app_version_public_id = active_version['public_id']

        logging.info(f'Loaded package: {self.app_author}/{self.app_name}')

        cwd = os.getcwd()

        if stdin is None:
            stdin = b''

        if not output_path.startswith('/'):
            # output_path is relative, make absolute
            output_path = f'{cwd}/{output_path}'

        if isinstance(args, str):
            args = list(filter(lambda p: p != '', args.split(' ')))

        if not isinstance(args, list):
            raise Exception('The given input arguments must be list or str')

        if isinstance(stdin, str):
            stdin = stdin.encode('utf-8')

        if files is None:
            files = []
            for idx, arg in enumerate(args):
                if os.path.isfile(arg):
                    files.append(arg)
                    args[idx] = arg.split('/')[-1]

        files_dict = {}

        for file in files:
            path = file
            if not file.startswith('/'):
                # make path absolute
                path = cwd + '/' + file

            arg_split = path.split('/')
            file = open(path, 'rb')
            path = '/' + arg_split[-1]

            files_dict[path] = file.read()
            file.close()

        BioLibServer.log_level = BioLib.log_level
        with BioLibServer(('127.0.0.1', 0), BioLibHandler) as httpd:
            port = httpd.server_address[1]
            thread = threading.Thread(target=httpd.serve_forever)
            # TODO: figure out how we can avoid changing the current directory
            os.chdir(os.path.dirname(os.path.realpath(__file__)) + '/biolib-js/')
            try:
                thread.start()
                res = asyncio.get_event_loop().run_until_complete(
                    self.call_pyppeteer(port=port, args=args, stdin=stdin, files=files_dict,
                                        output_dir_path=output_path))
            finally:
                os.chdir(cwd)
                httpd.shutdown()
                thread.join()
            return res

        raise BioLibError('Failed to start TCPServer')



import subprocess
import threading
import time
import tarfile
import zipfile
import logging
import os
import io

import docker
from docker.errors import ImageNotFound, APIError

from biolib.compute_node.compute_process import compute_process_config
from biolib.compute_node.compute_process.mappings import Mappings, path_without_first_folder
from biolib.compute_node.compute_process.utils import ComputeProcessException
from biolib.compute_node.utils import SystemExceptionCodes, random_string


class BiolibDockerClient():
    """
    An abstraction to manage and run biolib docker containers
    """

    def __init__(self, is_running_in_enclave, messages_to_send_queue):
        self.is_running_in_enclave = is_running_in_enclave
        self.messages_to_send_queue = messages_to_send_queue
        self.biolib_logger = logging.getLogger('biolib')
        self.docker_api_client = docker.APIClient(base_url=compute_process_config.DOCKER_SOCKET_PATH)
        self.docker_client = docker.from_env()
        self.container = None
        self.container_id = None
        self.source_mappings = None
        self.input_mappings = None
        self.output_mappings = None
        self.arguments = None
        self.biolib_logger = logging.getLogger('biolib')
        self.proxies = []
        self.remote_hostnames = []
        self.random_docker_id = random_string(15)
        self.internal_network = None
        self.public_network = None
        self.compute_process_dir = os.path.dirname(os.path.realpath(__file__))
        self.runtime_tar_path = f'{self.compute_process_dir}/tars/runtime_{self.random_docker_id}.tar'
        self.input_tar_path = f'{self.compute_process_dir}/tars/input_{self.random_docker_id}.tar'

    def delete_old_state(self):
        tar_time = time.time()
        for path_to_delete in [self.input_tar_path, self.runtime_tar_path]:
            if os.path.exists(path_to_delete):
                os.remove(path_to_delete)
        self.biolib_logger.debug(f"Deleted tars in: {time.time() - tar_time}")

        proxy_time = time.time()
        for proxy in self.proxies:
            proxy['container'].remove(force=True)
        self.biolib_logger.debug(f"Deleted proxies in: {time.time() - proxy_time}")

        container_time = time.time()
        if self.container:
            if self.container.status != 'exited':
                self.container.stop()
            self.container.remove()
        self.biolib_logger.debug(f"Deleted compute container in: {time.time() - container_time}")

        network_time = time.time()
        if self.internal_network:
            self.internal_network.remove()
        if self.public_network:
            self.public_network.remove()
        self.biolib_logger.debug(f"Deleted networks in: {time.time() - network_time}")

    def create_networks(self):
        try:
            self.internal_network = self.docker_client.networks.create(
                name='biolib-sandboxed-network-' + self.random_docker_id,
                internal=True,
                driver="bridge"
            )
            self.public_network = self.docker_client.networks.create(
                name='biolib-proxy-network-' + self.random_docker_id,
                internal=False,
                driver="bridge"
            )
        except Exception as exception:
            raise ComputeProcessException(exception, SystemExceptionCodes.FAILED_TO_CREATE_DOCKER_NETWORKS.value,
                                          self.messages_to_send_queue) from exception

    def start_remote_host_proxies(self, remote_hostnames, remote_host_lock: threading.Lock):
        try:
            remote_host_lock.acquire()

            # The enclave currently always has the remote-host-proxy image
            if not self.is_running_in_enclave:
                try:
                    self.docker_client.images.get('biolib/remote-host-proxy:latest')
                except ImageNotFound:
                    self.biolib_logger.debug('Pulling remote host docker image...')
                    self.docker_client.images.pull('biolib/remote-host-proxy:latest')

            nr_of_desired_remote_hosts = len(self.proxies) + len(remote_hostnames)
            threading.Thread(target=self._release_remote_hosts_proxies_lock_when_all_created,
                             args=(nr_of_desired_remote_hosts, remote_host_lock)).start()

            # This allows the function to be called multiple times with different hosts in an enclave
            last_used_vsock_port = 8000 + len(self.proxies)
            last_used_local_port = 64000 + len(self.proxies)

            for host_id, hostname in enumerate(remote_hostnames):
                if self.is_running_in_enclave:
                    vsock_port = last_used_vsock_port + host_id
                    local_port = last_used_local_port + host_id
                    forwarder_cid = '3'
                    self.biolib_logger.debug(f"Starting traffic forwarder forwarding to port {local_port} "
                                             f"for hostname {hostname}")
                    subprocess.Popen(['python3', f'{self.compute_process_dir}/traffic_forwarder.py', str(local_port),
                                      forwarder_cid, str(vsock_port)])
                else:
                    local_port = None

                threading.Thread(target=self._start_remote_host_proxy, args=(hostname, local_port)).start()

        except Exception as exception:
            raise ComputeProcessException(exception, SystemExceptionCodes.FAILED_TO_START_REMOTE_HOST_PROXIES.value,
                                          self.messages_to_send_queue) from exception

    def _release_remote_hosts_proxies_lock_when_all_created(self, nr_of_desired_remote_hosts,
                                                            remote_host_lock: threading.Lock):
        # Block until all remote hosts proxies has been created
        # TODO: Maybe write this with locks
        start_time = time.time()
        while len(self.proxies) != nr_of_desired_remote_hosts:
            time.sleep(0.1)
            continue

        try:
            if self.is_running_in_enclave:
                hosts_file = open('/etc/hosts', 'a')
                for proxy in self.proxies:
                    hosts_file.write(f'\n{proxy["ip"]} {proxy["hostname"]}')
                hosts_file.close()
            remote_host_lock.release()
        except Exception as exception:
            raise ComputeProcessException(exception,
                                          SystemExceptionCodes.FAILED_TO_REDIRECT_ENCLAVE_TRAFFIC_TO_PROXIES.value,
                                          self.messages_to_send_queue) from exception

        self.biolib_logger.debug(f'Starting remote hosts proxies took: {time.time() - start_time}s')

    def _start_remote_host_proxy(self, remote_hostname, local_traffic_forwarder_port=None):
        try:
            if self.is_running_in_enclave:
                extra_hosts = {remote_hostname: self.public_network.attrs['IPAM']['Config'][0]['Gateway']}
            else:
                extra_hosts = None

            proxy_container = self.docker_client.containers.create(
                image='biolib/remote-host-proxy:latest',
                network=self.internal_network.name,
                publish_all_ports=True,
                name='biolib-remote-hosts-proxy-' + remote_hostname + '-' + self.random_docker_id,
                extra_hosts=extra_hosts,
                detach=True
            )
        except Exception as exception:
            raise ComputeProcessException(exception, SystemExceptionCodes.FAILED_TO_CREATE_PROXY_CONTAINER.value,
                                          self.messages_to_send_queue) from exception

        try:
            if self.is_running_in_enclave:
                upstream_http_port = local_traffic_forwarder_port
                upstream_https_port = local_traffic_forwarder_port
            else:
                upstream_http_port = 80
                upstream_https_port = 443

            nginx_config = f"""
                events {{}}
                error_log /dev/stdout info;
                stream {{
                    upstream forward-http {{
                        server {remote_hostname}:{upstream_http_port};
                    }}
                    upstream forward-https {{
                        server {remote_hostname}:{upstream_https_port};
                    }}
                    server {{
                        listen     80;
                        proxy_pass forward-http;
                    }}
                    server {{
                        listen     443;
                        proxy_pass forward-https;
                    }}
                }}
            """
            tar_name = f'{self.compute_process_dir}/tars/remote_hosts_{self.random_docker_id}_{remote_hostname}.tar'
            remote_host_tar = tarfile.open(tar_name, 'w')
            nginx_conf_path = '/etc/nginx/nginx.conf'
            self.add_file_to_tar(remote_host_tar, nginx_conf_path, nginx_conf_path, nginx_config.encode())
            remote_host_tar.close()
            remote_hosts_tar_data = open(tar_name, 'rb').read()
            self.docker_api_client.put_archive(proxy_container.id, '/', remote_hosts_tar_data)
        except Exception as exception:
            raise ComputeProcessException(exception, SystemExceptionCodes.FAILED_TO_CONFIGURE_ALLOWED_REMOTE_HOST.value,
                                          self.messages_to_send_queue) from exception

        try:
            self.docker_api_client.start(proxy_container.id)
            self.public_network.connect(proxy_container.id)

            proxy_container.reload()
            proxy_container_ip = proxy_container.attrs['NetworkSettings']['Networks'][self.internal_network.name][
                'IPAddress']
            self.proxies.append({
                'container': proxy_container,
                'ip': proxy_container_ip,
                'hostname': remote_hostname,
            })
            os.remove(tar_name)
        except Exception as exception:
            raise ComputeProcessException(exception, SystemExceptionCodes.FAILED_TO_CONFIGURE_ALLOWED_REMOTE_HOST.value,
                                          self.messages_to_send_queue) from exception

    def pull_image(self, image_uri, access_token, job_id, image_lock: threading.Lock(), ecr_proxy, enclave_ecr_token):
        try:
            image_lock.acquire()
            threading.Thread(target=self._pull, args=(image_uri, access_token, job_id, image_lock, ecr_proxy,
                                                      enclave_ecr_token)).start()
        except Exception as exception:
            raise ComputeProcessException(exception, SystemExceptionCodes.FAILED_TO_START_IMAGE_PULLING_THREAD.value,
                                          self.messages_to_send_queue) from exception

    def _pull(self, image_uri, access_token, job_id, image_lock, ecr_proxy, enclave_ecr_token):
        try:
            start_time = time.time()
            image_uri = f'{ecr_proxy}/{image_uri}'

            try:
                self.docker_client.images.get(image_uri)
            except ImageNotFound:
                if enclave_ecr_token:
                    tokens = f'{enclave_ecr_token},{job_id}'
                    auth_config = {'username': 'AWS', 'password': tokens}
                else:
                    tokens = f'{access_token},{job_id}'
                    auth_config = {'username': 'AWS', 'password': tokens}
                self.docker_client.images.pull(image_uri, auth_config=auth_config)

            self.biolib_logger.debug(f"Pulled image in: {time.time() - start_time}")
            image_lock.release()
        except Exception as exception:
            raise ComputeProcessException(exception, SystemExceptionCodes.FAILED_TO_PULL_DOCKER_IMAGE.value,
                                          self.messages_to_send_queue) from exception

    def set_mappings(self, input_mappings, source_mappings, output_mappings, arguments):
        self.input_mappings = input_mappings
        self.source_mappings = source_mappings
        self.output_mappings = output_mappings
        self.arguments = arguments

    def create_compute_container(self, image, command, working_dir):
        try:
            if self.proxies:
                extra_hosts = {proxy['hostname']: proxy['ip'] for proxy in self.proxies}
            else:
                extra_hosts = {}

            self.container = self.docker_client.containers.create(
                image=image,
                command=command,
                working_dir=working_dir,
                network=self.internal_network.name,
                extra_hosts=extra_hosts
            )
            self.container_id = self.container.id
        except Exception as exception:
            raise ComputeProcessException(exception,
                                          SystemExceptionCodes.FAILED_TO_START_COMPUTE_CONTAINER.value,
                                          self.messages_to_send_queue) from exception

    def run_container(self):
        try:
            self.docker_api_client.start(self.container_id)

            exit_code = self.docker_api_client.wait(self.container_id)['StatusCode']

            stdout = self.docker_api_client.logs(self.container_id, stdout=True, stderr=False)
            stderr = self.docker_api_client.logs(self.container_id, stdout=False, stderr=True)

            mapped_output_files = self.get_output_files()
            return stdout, stderr, exit_code, mapped_output_files

        except Exception as exception:
            raise ComputeProcessException(exception, SystemExceptionCodes.FAILED_TO_RUN_COMPUTE_CONTAINER.value,
                                          self.messages_to_send_queue) from exception

    def add_file_to_tar(self, tar, current_path, mapped_path, data):
        if current_path.endswith('/'):
            # Remove trailing slash as tarfile.addfile appends it automatically
            tarinfo = tarfile.TarInfo(name=mapped_path[:-1])
            tarinfo.type = tarfile.DIRTYPE
            tar.addfile(tarinfo)

        else:
            tarinfo = tarfile.TarInfo(name=mapped_path)
            file_like = io.BytesIO(data)
            tarinfo.size = len(file_like.getvalue())
            tar.addfile(tarinfo, file_like)

    def make_input_tar(self, files):
        input_tar = tarfile.open(self.input_tar_path, 'w')
        input_mappings = Mappings(self.input_mappings, self.arguments)
        for path, data in files.items():
            # Make all paths absolute
            if not path.startswith('/'):
                path = '/' + path

            mapped_file_names = input_mappings.get_mappings_for_path(path)
            for mapped_file_name in mapped_file_names:
                self.add_file_to_tar(tar=input_tar, current_path=path, mapped_path=mapped_file_name, data=data)

        input_tar.close()

    def make_runtime_tar(self, runtime_zip_data):
        runtime_tar = tarfile.open(self.runtime_tar_path, 'w')
        runtime_zip = zipfile.ZipFile(io.BytesIO(runtime_zip_data))
        source_mappings = Mappings(self.source_mappings, self.arguments)

        for zip_file_name in runtime_zip.namelist():
            # Make paths absolute and remove root folder from path
            file_path = '/' + path_without_first_folder(zip_file_name)
            mapped_file_names = source_mappings.get_mappings_for_path(file_path)
            for mapped_file_name in mapped_file_names:
                file_data = runtime_zip.read(zip_file_name)
                self.add_file_to_tar(tar=runtime_tar, current_path=zip_file_name, mapped_path=mapped_file_name,
                                     data=file_data)

        runtime_tar.close()

    def map_and_copy_input_files_to_container(self, files):
        try:
            self.make_input_tar(files)
            input_tar_bytes = open(self.input_tar_path, 'rb').read()
            self.docker_api_client.put_archive(self.container_id, '/', input_tar_bytes)
        except Exception as exception:
            raise ComputeProcessException(exception,
                                          SystemExceptionCodes.FAILED_TO_COPY_INPUT_FILES_TO_COMPUTE_CONTAINER.value,
                                          self.messages_to_send_queue) from exception

    def map_and_copy_runtime_files_to_container(self, runtime_zip_data):
        try:
            self.make_runtime_tar(runtime_zip_data)
            runtime_tar_bytes = open(self.runtime_tar_path, 'rb').read()
            self.docker_api_client.put_archive(self.container_id, '/', runtime_tar_bytes)
        except Exception as exception:
            raise ComputeProcessException(exception,
                                          SystemExceptionCodes.FAILED_TO_COPY_RUNTIME_FILES_TO_COMPUTE_CONTAINER.value,
                                          self.messages_to_send_queue) from exception

    def get_output_files(self):
        try:
            input_tar = tarfile.open(self.input_tar_path)
            input_tar_filelist = input_tar.getnames()

            if os.path.exists(self.runtime_tar_path):
                runtime_tar = tarfile.open(self.runtime_tar_path)
                runtime_tar_filelist = runtime_tar.getnames()
            else:
                runtime_tar = None

            mapped_output_files = {}
            for mapping in self.output_mappings:
                try:
                    tar_bytes_generator, _ = self.docker_api_client.get_archive(self.container_id, mapping['from_path'])
                except APIError:
                    self.biolib_logger.warning(f'Could not get output from path {mapping["from_path"]}')
                    continue

                tar_bytes_obj = io.BytesIO()
                for chunk in tar_bytes_generator:
                    tar_bytes_obj.write(chunk)

                tar = tarfile.open(fileobj=io.BytesIO(tar_bytes_obj.getvalue()))
                for file in tar.getmembers():
                    file_obj = tar.extractfile(file)

                    # Skip empty dirs
                    if not file_obj:
                        continue
                    file_data = tar.extractfile(file).read()

                    # Remove parent dir from tar file name and prepend from_path.
                    # Except if from_path is root '/', that works out of the box
                    if mapping['from_path'].endswith('/') and mapping['from_path'] != '/':
                        file_name = mapping['from_path'] + path_without_first_folder(file.name)

                    # When getting a file use the from_path.
                    # This is due to directory info (absolute path) being lost when using get_archive on files
                    else:
                        file_name = mapping['from_path']

                    # Filter out unchanged input files
                    if file_name in input_tar_filelist and input_tar.extractfile(file_name).read() == file_data:
                        continue

                    # Filter out unchanged source files if provided
                    if runtime_tar and file_name in runtime_tar_filelist and runtime_tar.extractfile(
                            file_name).read() == file_data:
                        continue

                    mapped_file_names = Mappings([mapping], self.arguments).get_mappings_for_path(file_name)
                    for mapped_file_name in mapped_file_names:
                        mapped_output_files[mapped_file_name] = file_data

        except Exception as exception:
            raise ComputeProcessException(exception, SystemExceptionCodes.FAILED_TO_RUN_COMPUTE_CONTAINER.value,
                                          self.messages_to_send_queue) from exception

        return mapped_output_files



import logging
import random
import subprocess
import sys
import threading
import socket
from queue import Queue

from biolib.compute_node.socker_listener_thread import SocketListenerThread
from biolib.compute_node.socket_sender_thread import SocketSenderThread
from biolib.biolib_binary_format import AttestationDocument, SystemStatusUpdate, SystemException
from biolib.compute_node.utils import get_package_type, WorkerThreadException, SystemExceptionCodes
from biolib.compute_node.compute_process import compute_process_config

SOCKET_HOST = '127.0.0.1'


class WorkerThread(threading.Thread):
    def __init__(self, compute_state, eif_path=None):
        try:
            super().__init__()
            self.compute_state = compute_state
            self.eif_path = eif_path
            if self.eif_path:
                self.socket_port = 5005
            else:
                self.socket_port = random.choice(range(6000, 65000))
            self.socket = None
            self.connection = None
            self.compute_process = None
            self.connection_thread = None
            self.listener_thread = None
            self.sender_thread = None
            self.biolib_logger = logging.getLogger('biolib')
            self._start_and_connect_to_compute_process()

            self.biolib_logger.debug(f"WorkerThread connected to port {self.socket_port}")

        except Exception as exception:
            raise WorkerThreadException(exception, SystemExceptionCodes.FAILED_TO_INITIALIZE_WORKER_THREAD.value,
                                        worker_thread=self) from exception

    def run(self):
        try:
            while True:
                package = self.compute_state['received_messages_queue'].get()
                package_type = get_package_type(package)

                if package_type == 'AttestationDocument':
                    self.compute_state['attestation_document'] = AttestationDocument(package).deserialize()

                elif package_type == 'SystemStatusUpdate':
                    progress, log_message = SystemStatusUpdate(package).deserialize()
                    self.compute_state['status']['status_updates'].append({'progress': progress,
                                                                           'log_message': log_message})

                elif package_type == 'SystemException':
                    error_code = SystemException(package).deserialize()
                    self.compute_state['status']['error_code'] = error_code
                    self.biolib_logger.debug("Hit error. Terminating Worker Thread and Compute Process")
                    self.terminate()

                elif package_type == 'ModuleOutput' or package_type == 'AesEncryptedPackage':
                    self.compute_state['result'] = package
                    self.compute_state['status']['status_updates'].append({'progress': 95,
                                                                           'log_message': 'Result Ready'})
                    self.terminate()

                else:
                    raise Exception(f'Package type from child was not recognized: {package}')

                self.compute_state['received_messages_queue'].task_done()

        except Exception as exception:
            raise WorkerThreadException(exception, SystemExceptionCodes.FAILED_TO_HANDLE_PACKAGE_IN_WORKER_THREAD.value,
                                        worker_thread=self) from exception

    def _start_and_connect_to_compute_process(self):
        if self.eif_path:
            self.socket = socket.socket(socket.AF_VSOCK, socket.SOCK_STREAM)  # pylint: disable=no-member
            cid = socket.VMADDR_CID_ANY  # pylint: disable=no-member
            self.socket.bind((cid, self.socket_port))
        else:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.socket.bind((SOCKET_HOST, self.socket_port))

        self.socket.listen()

        received_messages_queue = Queue()
        messages_to_send_queue = Queue()

        # Starting a thread for accepting connections before starting the process that should to connect to the socket
        self.connection_thread = threading.Thread(target=self._accept_new_socket_connection, args=[
            received_messages_queue,
            messages_to_send_queue
        ])
        self.connection_thread.start()

        if self.eif_path:
            self.compute_process = subprocess.Popen(['sudo', 'nitro-cli', 'run-enclave',
                                                     '--cpu-count',
                                                     compute_process_config.CPU_COUNT_TO_PROVISION_ENCLAVE,
                                                     '--memory',
                                                     compute_process_config.MEMORY_TO_PROVISION_ENCLAVE,
                                                     '--eif-path', self.eif_path])

        else:
            self.compute_process = subprocess.Popen(['biolib', 'start-compute-process', '--port',
                                                     str(self.socket_port)])

        self.compute_state['received_messages_queue'] = received_messages_queue
        self.compute_state['messages_to_send_queue'] = messages_to_send_queue
        self.compute_state['worker_thread'] = self

    def _accept_new_socket_connection(self, received_messages_queue, messages_to_send_queue):
        self.connection, _ = self.socket.accept()
        self.listener_thread = SocketListenerThread(self.connection, received_messages_queue)
        self.listener_thread.start()

        self.sender_thread = SocketSenderThread(self.connection, messages_to_send_queue)
        self.sender_thread.start()

    def terminate(self):
        if self.compute_process:
            self.biolib_logger.debug(f"Terminating Compute Process with PID {self.compute_process.pid}")
            self.compute_state['messages_to_send_queue'].put(b'STOP')

        if self.sender_thread:
            self.sender_thread.terminate()

        if self.listener_thread:
            self.listener_thread.terminate()

        if self.socket:
            self.socket.close()

        if self.connection:
            self.connection.close()

        self.biolib_logger.debug("Terminating Worker Thread")
        sys.exit()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Browser module."""
import asyncio
import logging
from asyncio import Future
from subprocess import Popen
from typing import TYPE_CHECKING, Awaitable, Callable, Dict, List, Optional, Sequence

from pyee import AsyncIOEventEmitter
from biolib.pyppeteer.pyppeteer.connection import Connection
from biolib.pyppeteer.pyppeteer.errors import BrowserError
from biolib.pyppeteer.pyppeteer.events import Events
from biolib.pyppeteer.pyppeteer.models import Protocol, WebPermission
from biolib.pyppeteer.pyppeteer.target import Target
from biolib.pyppeteer.pyppeteer.task_queue import TaskQueue

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from pyppeteer.page import Page


class Browser(AsyncIOEventEmitter):
    """Browser class.

    A Browser object is created when pyppeteer connects to chrome, either
    through :func:`~pyppeteer.launcher.launch` or
    :func:`~pyppeteer.launcher.connect`.
    """

    def __init__(
        self,
        connection: Connection,
        contextIds: List[str],
        ignoreHTTPSErrors: bool,
        defaultViewport: Protocol.Page.Viewport,
        process: Optional[Popen] = None,
        closeCallback: Callable[[], Awaitable[None]] = None,
    ) -> None:
        super().__init__()
        self._ignoreHTTPSErrors = ignoreHTTPSErrors
        self._defaultViewport = defaultViewport
        self._process = process
        self._screenshotTaskQueue = TaskQueue()
        self._connection = connection
        self.loop = self._connection.loop

        if closeCallback:
            self._closeCallback = closeCallback
        else:

            async def _dummy_callback() -> None:
                pass

            self._closeCallback = _dummy_callback

        self._defaultContext = BrowserContext(self._connection, self, None)
        self._contexts: Dict[str, BrowserContext] = {}
        for contextId in contextIds:
            self._contexts[contextId] = BrowserContext(self._connection, self, None)

        self._targets: Dict[str, Target] = {}
        self._connection.on(Events.Connection.Disconnected, lambda: self.emit(Events.Browser.Disconnected))
        self._connection.on(
            'Target.targetCreated', lambda event: self.loop.create_task(self._targetCreated(event)),
        )
        self._connection.on(
            'Target.targetDestroyed', lambda event: self.loop.create_task(self._targetDestroyed(event)),
        )
        self._connection.on(
            'Target.targetInfoChanged', lambda event: self.loop.create_task(self._targetInfoChanged(event)),
        )

    @property
    def process(self) -> Optional[Popen]:
        """Return process of this browser.

        If browser instance is created by :func:`pyppeteer.launcher.connect`,
        return ``None``.
        """
        return self._process

    async def createIncognitoBrowserContext(self) -> 'BrowserContext':
        """Create a new incognito browser context.

        This won't share cookies/cache with other browser contexts.

        .. code::

            browser = await launch()
            # Create a new incognito browser context.
            context = await browser.createIncognitoBrowserContext()
            # Create a new page in a pristine context.
            page = await context.newPage()
            # Do stuff
            await page.goto('https://example.com')
            ...
        """
        obj = await self._connection.send('Target.createBrowserContext')
        browserContextId = obj['browserContextId']
        context = BrowserContext(self._connection, self, browserContextId)
        self._contexts[browserContextId] = context
        return context

    @property
    def browserContexts(self) -> List['BrowserContext']:
        """Return a list of all open browser contexts.

        In a newly created browser, this will return a single instance of
        ``[BrowserContext]``
        """
        return [self._defaultContext] + [context for context in self._contexts.values()]

    @property
    def defaultBrowserContext(self) -> 'BrowserContext':
        return self._defaultContext

    async def _disposeContext(self, contextId: str) -> None:
        await self._connection.send('Target.disposeBrowserContext', {'browserContextId': contextId,})
        self._contexts.pop(contextId, None)

    @staticmethod
    async def create(
        connection: Connection,
        contextIds: List[str],
        ignoreHTTPSErrors: bool,
        defaultViewport: Protocol.Page.Viewport,
        process: Optional[Popen] = None,
        closeCallback: Callable[[], Awaitable[None]] = None,
    ) -> 'Browser':
        """Create browser object."""
        browser = Browser(connection, contextIds, ignoreHTTPSErrors, defaultViewport, process, closeCallback)
        await connection.send('Target.setDiscoverTargets', {'discover': True})
        return browser

    async def _targetCreated(self, event: Dict) -> None:
        targetInfo = event['targetInfo']
        browserContextId = targetInfo.get('browserContextId')

        if browserContextId and browserContextId in self._contexts:
            context = self._contexts[browserContextId]
        else:
            context = self._defaultContext

        target = Target(
            targetInfo=targetInfo,
            browserContext=context,
            sessionFactory=lambda: self._connection.createSession(targetInfo),
            ignoreHTTPSErrors=self._ignoreHTTPSErrors,
            defaultViewport=self._defaultViewport,
            screenshotTaskQueue=self._screenshotTaskQueue,
            loop=self._connection.loop,
        )
        if targetInfo['targetId'] in self._targets:
            raise BrowserError('target should not exist before create.')
        self._targets[targetInfo['targetId']] = target
        if await target._initializedPromise:
            self.emit(Events.Browser.TargetCreated, target)
            context.emit(Events.BrowserContext.TargetCreated, target)

    async def _targetDestroyed(self, event: Dict) -> None:
        target = self._targets[event['targetId']]
        del self._targets[event['targetId']]
        target._closedCallback()
        if await target._initializedPromise:
            self.emit(Events.Browser.TargetDestroyed, target)
            target.browserContext.emit(Events.BrowserContext.TargetDestroyed, target)
        target._initializedCallback(False)

    async def _targetInfoChanged(self, event: Dict) -> None:
        target = self._targets.get(event['targetInfo']['targetId'])
        if not target:
            raise BrowserError('target should exist before targetInfoChanged')
        previousURL = target.url
        wasInitialized = target._isInitialized
        target._targetInfoChanged(event['targetInfo'])
        if wasInitialized and previousURL != target.url:
            self.emit(Events.Browser.TargetChanged, target)
            target.browserContext.emit(Events.BrowserContext.TargetChanged, target)

    @property
    def wsEndpoint(self) -> str:
        """Return websocket end point url."""
        return self._connection.url

    async def newPage(self) -> 'Page':
        """Make new page on this browser and return its object."""
        return await self._defaultContext.newPage()

    async def _createPageInContext(self, contextId: Optional[str]) -> 'Page':
        options = {'url': 'about:blank'}
        if contextId:
            options['browserContextId'] = contextId

        targetId = (await self._connection.send('Target.createTarget', options)).get('targetId')
        target = self._targets.get(targetId)
        if target is None or not await target._initializedPromise:
            raise BrowserError('Failed to create target for page.')
        page = await target.page()
        if page is None:
            raise BrowserError('Failed to create page.')
        return page

    def targets(self) -> List[Target]:
        """Get a list of all active targets inside the browser.

        In case of multiple browser contexts, this will return a list
        with all the targets in all browser contexts.
        """
        return [target for target in self._targets.values() if target._isInitialized]

    @property
    def target(self) -> Target:
        """get active browser target"""
        return next((target for target in self.targets() if target.type == 'browser'))

    async def waitForTarget(self, predicate: Callable[[Target], bool], timeout: float = 30_000) -> Target:
        """
        Wait for target that matches predicate function.
        :param predicate: function that takes 1 argument of Target object
        :param timeout: how long to wait for target in milliseconds,
        TimeoutError will be raised otherwise
        """
        if timeout:  # js uses ms while asyncio uses seconds
            timeout = timeout / 1_000
        existing_target = [target for target in self.targets() if predicate(target)]
        if existing_target:
            return existing_target[0]

        result_fut: Future[Target] = self.loop.create_future()

        def check(target: Target) -> None:
            if predicate(target):
                result_fut.set_result(target)

        self.on(Events.Browser.TargetCreated, check)
        self.on(Events.Browser.TargetChanged, check)
        result = await asyncio.wait_for(result_fut, timeout=timeout)
        self.remove_listener(Events.Browser.TargetCreated, check)
        self.remove_listener(Events.Browser.TargetChanged, check)
        return result

    @property
    async def pages(self) -> List['Page']:
        """Get all pages of this browser.

        Non visible pages, such as ``"background_page"``, will not be listed
        here. You can find then using :meth:`pyppeteer.target.Target.page`.

        In case of multiple browser contexts, this method will return a list
        with all the pages in all browser contexts.
        """
        pages = await asyncio.gather(*[context.pages() for context in self.browserContexts])
        return [p for ps in pages for p in ps]

    async def version(self) -> str:
        """Get version of the browser."""
        version = await self._getVersion()
        return version['product']

    async def userAgent(self) -> str:
        """Return browser's original user agent.

        .. note::
            Pages can override browser user agent with
            :meth:`pyppeteer.page.Page.setUserAgent`.
        """
        version = await self._getVersion()
        return version.get('userAgent', '')

    async def close(self) -> None:
        """Close connections and terminate browser process."""
        await self._closeCallback()
        await self.disconnect()

    async def disconnect(self) -> None:
        """Disconnect browser."""
        await self._connection.dispose()

    @property
    def isConnected(self) -> bool:
        return not self._connection._closed

    def _getVersion(self) -> Awaitable:
        return self._connection.send('Browser.getVersion')


class BrowserContext(AsyncIOEventEmitter):
    """BrowserContext provides multiple independent browser sessions.

    When a browser is launched, it has a single BrowserContext used by default.
    The method `browser.newPage()` creates a page in the default browser
    context.

    If a page opens another page, e.g. with a ``window.open`` call, the popup
    will belong to the parent page's browser context.

    Pyppeteer allows creation of "incognito" browser context with
    ``browser.createIncognitoBrowserContext()`` method.
    "incognito" browser contexts don't write any browser data to disk.

    .. code::

        # Create new incognito browser context
        context = await browser.createIncognitoBrowserContext()
        # Create a new page inside context
        page = await context.newPage()
        # ... do stuff with page ...
        await page.goto('https://example.com')
        # Dispose context once it's no longer needed
        await context.close()
    """

    def __init__(self, connection: Connection, browser: Browser, contextId: Optional[str]) -> None:
        super().__init__()
        self._connection = connection
        self._browser = browser
        self._id = contextId

    def targets(self) -> List[Target]:
        """Return a list of all active targets inside the browser context."""
        targets = []
        for target in self._browser.targets():
            if target.browserContext == self:
                targets.append(target)
        return targets

    async def pages(self) -> List['Page']:
        """Return list of all open pages.

        Non-visible pages, such as ``"background_page"``, will not be listed
        here. You can find them using :meth:`pyppeteer.target.Target.page`.
        """
        pages = [target.page() for target in self.targets() if target.type == 'page']
        return [page for page in await asyncio.gather(*pages) if page]

    def isIncognito(self) -> bool:
        """Return whether BrowserContext is incognito.

        The default browser context is the only non-incognito browser context.

        .. note::
            The default browser context cannot be closed.
        """
        return bool(self._id)

    async def overridePermissions(self, origin: str, permissions: Sequence[WebPermission]) -> None:
        web_perm_to_protocol = {
            'geolocation': 'geolocation',
            'midi': 'midi',
            'notifications': 'notifications',
            'push': 'push',
            'camera': 'videoCapture',
            'microphone': 'audioCapture',
            'background-sync': 'backgroundSync',
            'ambient-light-sensor': 'sensors',
            'accelerometer': 'sensors',
            'gyroscope': 'sensors',
            'magnetometer': 'sensors',
            'accessibility-events': 'accessibilityEvents',
            'clipboard-read': 'clipboardRead',
            'clipboard-write': 'clipboardWrite',
            'payment-handler': 'paymentHandler',
            # chrome specific
            'midi-sysex': 'midiSysex',
        }
        protocol_perms = []
        for perm in permissions:
            protocol_perm = web_perm_to_protocol.get(perm)
            if protocol_perm is None:
                raise RuntimeError(f'Unknown permission: {perm}')
            protocol_perms.append(perm)
        await self._connection.send(
            'Browser.grantPermissions', {'origin': origin, 'browserContextId': self._id, 'permissions': permissions}
        )

    async def clearPermissionOverrides(self) -> None:
        await self._connection.send('Browser.resetPermissions', {'browserContextId': self._id})

    async def newPage(self) -> 'Page':
        """Create a new page in the browser context."""
        return await self._browser._createPageInContext(self._id)

    @property
    def browser(self) -> Browser:
        """Return the browser this browser context belongs to."""
        return self._browser

    async def close(self) -> None:
        """Close the browser context.

        All the targets that belongs to the browser context will be closed.

        .. note::
            Only incognito browser context can be closed.
        """
        if self._id is None:
            raise BrowserError('Non-incognito profile cannot be closed')
        await self._browser._disposeContext(self._id)



import socket
import sys
import threading
import time


def server(local_port, remote_cid, remote_port):
    dock_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    dock_socket.bind(('', local_port))
    dock_socket.listen(5)

    while True:
        client_socket, _ = dock_socket.accept()

        server_socket = socket.socket(socket.AF_VSOCK, socket.SOCK_STREAM)  # pylint: disable=no-member
        server_socket.connect((remote_cid, remote_port))

        outgoing_thread = threading.Thread(target=forward, args=(client_socket, server_socket))
        incoming_thread = threading.Thread(target=forward, args=(server_socket, client_socket))

        outgoing_thread.start()
        incoming_thread.start()


def forward(source, destination):
    string = ' '
    while string:
        string = source.recv(1024)
        if string:
            destination.sendall(string)
        else:
            source.shutdown(socket.SHUT_RD)
            destination.shutdown(socket.SHUT_WR)


def main(args):
    local_port = int(args[0])
    remote_cid = int(args[1])
    remote_port = int(args[2])

    thread = threading.Thread(target=server, args=(local_port, remote_cid, remote_port))
    thread.start()

    while True:
        time.sleep(60)


if __name__ == '__main__':
    main(sys.argv[1:])



CLIENT_PORT = 5005
DOCKER_SOCKET_PATH = 'unix://var/run/docker.sock'

# The default recv buffer size
DEFAULT_BUFFER_SIZE = 32768

# The amount of RAM (in MB) to provision the enclave. Half of it will become disk space.
MEMORY_TO_PROVISION_ENCLAVE = '26624'

# The amount of vCPUs to provision the enclave. This has to be atleast 2.
CPU_COUNT_TO_PROVISION_ENCLAVE = '2'




# Time in minutes before a Compute Node always shuts down
COMPUTE_NODE_AUTO_SHUTDOWN_TIME_MINUTES = 1440 #  24 hours

# Time in minutes before a Compute Node running a job shuts down
COMPUTE_NODE_RUNNING_JOB_SHUTDOWN_TIME_MINUTES = 30

LOG_FILE_PATH = '/tmp/biolib.log'

GUNICORN_REQUEST_TIMEOUT = 300

VSOCK_PROXY_PATH = '/home/ec2-user/biolib-nitro-enclave-ami-setup/biolib-vsock-proxy'



from enum import Enum
import logging
import random
import string

biolib_logger = logging.getLogger('biolib')


def get_package_type(package):
    package_type = int.from_bytes(package[1:2], 'big')
    if package_type == 1:
        return 'ModuleInput'
    elif package_type == 2:
        return 'ModuleOutput'
    elif package_type == 3:  # Not used in the compute node
        return 'ModuleSource'
    elif package_type == 4:
        return 'AttestationDocument'
    elif package_type == 5:
        return 'SavedJob'
    elif package_type == 6:
        return 'RsaEncryptedAesPackage'
    elif package_type == 7:
        return 'AesEncryptedPackage'
    elif package_type == 8:
        return 'SystemStatusUpdate'
    elif package_type == 9:
        return 'SystemException'

    else:
        raise Exception(f"Unexpected package type {package_type}")


class SystemExceptionCodes(Enum):
    FAILED_TO_INIT_COMPUTE_PROCESS_VARIABLES = 1
    FAILED_TO_CONNECT_TO_WORKER_THREAD_SOCKET = 2
    FAILED_TO_START_SENDER_THREAD_OR_RECEIVER_THREAD = 3
    FAILED_TO_GET_ATTESTATION_DOCUMENT = 4
    FAILED_TO_CREATE_DOCKER_NETWORKS = 5
    FAILED_TO_START_REMOTE_HOST_PROXIES = 6
    FAILED_TO_REDIRECT_ENCLAVE_TRAFFIC_TO_PROXIES = 7
    FAILED_TO_CREATE_PROXY_CONTAINER = 8
    FAILED_TO_CONFIGURE_ALLOWED_REMOTE_HOST = 9
    FAILED_TO_SEND_STATUS_UPDATE = 10
    FAILED_TO_GET_REQUIRED_DATA_FOR_COMPUTE = 11
    FAILED_TO_START_RUNTIME_ZIP_DOWNLOAD_THREAD = 12
    FAILED_TO_DOWNLOAD_RUNTIME_ZIP = 13
    FAILED_TO_CONTACT_BACKEND_TO_CREATE_JOB = 14
    FAILED_TO_CREATE_NEW_JOB = 15
    FAILED_TO_START_IMAGE_PULLING_THREAD = 16
    FAILED_TO_PULL_DOCKER_IMAGE = 17
    FAILED_TO_START_COMPUTE_CONTAINER = 18
    FAILED_TO_COPY_INPUT_FILES_TO_COMPUTE_CONTAINER = 19
    FAILED_TO_COPY_RUNTIME_FILES_TO_COMPUTE_CONTAINER = 20
    FAILED_TO_RUN_COMPUTE_CONTAINER = 21
    FAILED_TO_RETRIEVE_AND_MAP_OUTPUT_FILES = 22
    FAILED_TO_SERIALIZE_AND_SEND_MODULE_OUTPUT = 23
    FAILED_TO_DESERIALIZE_SAVED_JOB = 24
    UNKOWN_COMPUTE_PROCESS_ERROR = 25
    FAILED_TO_INITIALIZE_WORKER_THREAD = 26
    FAILED_TO_HANDLE_PACKAGE_IN_WORKER_THREAD = 27


class WorkerThreadException(Exception):
    def __init__(self, original_error, error_code, worker_thread):
        super().__init__()
        worker_thread.compute_state['status']['error_code'] = error_code
        biolib_logger.error(original_error)
        worker_thread.terminate()


enclave_remote_hosts = ['biolib-cloud-api.s3.amazonaws.com',
                        'prod-eu-west-1-starport-layer-bucket.s3.eu-west-1.amazonaws.com',
                        'biolib.com',
                        'containers.biolib.com',
                        'staging-elb.biolib.com',
                        'containers.staging.biolib.com']


def random_string(length):
    return ''.join(random.choices(string.ascii_letters, k=length))



import logging
from queue import Queue

from biolib.biolib_binary_format import SystemException

biolib_logger = logging.getLogger('biolib')


class ComputeProcessException(Exception):
    def __init__(self, original_error: Exception, biolib_error_code, messages_to_send_queue: Queue):
        super().__init__()

        system_exception_package = SystemException().serialize(biolib_error_code)
        messages_to_send_queue.put(system_exception_package)

        biolib_logger.error(original_error)



import logging
import time
import subprocess
import json
import datetime
from socket import gethostname, gethostbyname
import os
import requests

from biolib.compute_node.worker_thread import WorkerThread
from biolib.compute_node import utils
from biolib.compute_node.webserver import webserver_config

biolib_logger = logging.getLogger('biolib')


def get_compute_state(unassigned_compute_processes):
    if len(unassigned_compute_processes) == 0:
        start_compute_process(unassigned_compute_processes)

    return unassigned_compute_processes.pop()


def start_compute_process(unassigned_compute_processes, eif_path=None):
    compute_state = {
        'status': {
            'status_updates': [
                {
                    'progress':  10,
                    'log_message': 'Initializing'
                }
            ],
        },
        'result': None,
        'attestation_document': b'',
        'received_messages_queue': None,
        'messages_to_send_queue': None,
        'worker_process': None
    }
    WorkerThread(compute_state, eif_path).start()
    while True:
        if compute_state['attestation_document']:
            break
        time.sleep(1)

    if eif_path:
        start_vsock_proxies(use_enclave_hosts=True)
        report_availability()

    unassigned_compute_processes.append(compute_state)


def start_vsock_proxies(remote_hosts=None, use_enclave_hosts=False):
    if use_enclave_hosts:
        remote_hosts = utils.enclave_remote_hosts
        baseport = 8000
    else:
        # Enclave remote hosts are already running, so we need to use next set of available ports
        baseport = 8000 + len(utils.enclave_remote_hosts)

    vsock_proxy_binary = webserver_config.VSOCK_PROXY_PATH
    log_file = open(webserver_config.LOG_FILE_PATH, 'a')
    for host_id, host in enumerate(remote_hosts):
        biolib_logger.debug(f"Opening vsock-proxy on port {baseport + host_id} for hostname {host}")
        subprocess.Popen([vsock_proxy_binary, str(baseport + host_id), host, '443'], stdout=log_file,
                         stderr=log_file)


def validate_saved_job(saved_job):
    if 'app_version' not in saved_job['job']:
        return False

    if 'modules' not in saved_job['job']['app_version']:
        return False

    if 'access_token' not in saved_job:
        return False

    if 'module_name' not in saved_job:
        return False

    return True


def report_availability():
    dev_mode = os.getenv('COMPUTE_NODE_ENV') == 'dev'
    if not dev_mode:
        try:
            if biolib_logger.isEnabledFor(logging.DEBUG):
                bash_rc_content = open('/home/ec2-user/.bashrc', 'r').read()
                biolib_logger.debug(f'.bashrc at the start of Compute Node: {bash_rc_content}')

            node_public_id = os.getenv('BIOLIB_COMPUTE_NODE_PUBLIC_ID')
            auth_token = os.getenv('BIOLIB_COMPUTE_NODE_AUTH_TOKEN')
            biolib_host = os.getenv('BIOLIB_HOST')
            ip_address = gethostbyname(gethostname())
            data = {'public_id':node_public_id, 'auth_token':auth_token, 'ip_address':ip_address}
            biolib_logger.debug(f'Registering with {data} to host {biolib_host}')
            biolib_logger.debug(f'Registering at {datetime.datetime.now()}')
            req = requests.post(f'{biolib_host}/api/jobs/report_available/', json=json.dumps(data))
            if req.status_code != 201:
                raise Exception("Non 201 error code")

            if req.json()['is_reserved']:
                # Start running job shutdown timer if reserved. It restarts when the job is actually saved
                subprocess.run(['sudo', 'shutdown', '-c'], check=True)
                subprocess.run(
                    ['sudo', 'shutdown', f'+{webserver_config.COMPUTE_NODE_RUNNING_JOB_SHUTDOWN_TIME_MINUTES}'],
                    check=True)

        except Exception as exception:  # pylint: disable=broad-except
            biolib_logger.error(f'Could not self register because of: {exception}')
            biolib_logger.debug("Self destructing")
            if not dev_mode:
                subprocess.run(['sudo', 'shutdown', 'now'], check=False)


def deregister():
    dev_mode = os.getenv('COMPUTE_NODE_ENV') == 'dev'
    if not dev_mode:
        node_public_id = os.getenv('BIOLIB_COMPUTE_NODE_PUBLIC_ID')
        auth_token = os.getenv('BIOLIB_COMPUTE_NODE_AUTH_TOKEN')
        biolib_host = os.getenv('BIOLIB_HOST')
        data = {'public_id': node_public_id, 'auth_token': auth_token}

        req = requests.post(f'{biolib_host}/api/jobs/deregister/', json=json.dumps(data))
        if req.status_code != 200:
            biolib_logger.error('Could not deregister!')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import asyncio
import logging
import re
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set

from pyee import AsyncIOEventEmitter
from biolib.pyppeteer.pyppeteer import helpers
from biolib.pyppeteer.pyppeteer.connection import CDPSession
from biolib.pyppeteer.pyppeteer.errors import BrowserError, ElementHandleError, PageError
from biolib.pyppeteer.pyppeteer.events import Events
from biolib.pyppeteer.pyppeteer.execution_context import ExecutionContext
from biolib.pyppeteer.pyppeteer.frame import Frame
from biolib.pyppeteer.pyppeteer.lifecycle_watcher import LifecycleWatcher
from biolib.pyppeteer.pyppeteer.models import WaitTargets
from biolib.pyppeteer.pyppeteer.network_manager import NetworkManager, Response
from biolib.pyppeteer.pyppeteer.timeout_settings import TimeoutSettings

if TYPE_CHECKING:
    from pyppeteer.page import Page

logger = logging.getLogger(__name__)

UTILITY_WORLD_NAME = '__pyppeteer_utility_world__'
EVALUATION_SCRIPT_URL = '__pyppeteer_evaluation_script__'
SOURCE_URL_REGEX = re.compile(r'^[ \t]*//[@#] sourceURL=\s*(\S*?)\s*$', re.MULTILINE,)


class FrameManager(AsyncIOEventEmitter):
    """FrameManager class."""

    def __init__(
        self, client: CDPSession, page: 'Page', ignoreHTTPSErrors: bool, timeoutSettings: TimeoutSettings
    ) -> None:
        """Make new frame manager."""
        super().__init__()
        self._client = client
        self._page = page
        self._networkManager = NetworkManager(client, ignoreHTTPSErrors, self)
        self._timeoutSettings = timeoutSettings
        self._mainFrame: Optional[Frame] = None
        self._frames: Dict[Any, Frame] = {}
        self._contextIdToContext: Dict[int, ExecutionContext] = {}
        self._isolatedWorlds: Set[str] = set()

        client.on(
            'Page.frameAttached',
            lambda event: self._onFrameAttached(event.get('frameId', ''), event.get('parentFrameId', '')),
        )
        client.on('Page.frameNavigated', lambda event: self._onFrameNavigated(event.get('frame')))
        client.on(
            'Page.navigatedWithinDocument',
            lambda event: self._onFrameNavigatedWithinDocument(event.get('frameId'), event.get('url')),
        )
        client.on('Page.frameDetached', lambda event: self._onFrameDetached(event.get('frameId')))
        client.on('Page.frameStoppedLoading', lambda event: self._onFrameStoppedLoading(event.get('frameId')))
        client.on(
            'Runtime.executionContextCreated', lambda event: self._onExecutionContextCreated(event.get('context'))
        )
        client.on(
            'Runtime.executionContextDestroyed',
            lambda event: self._onExecutionContextDestroyed(event.get('executionContextId')),
        )
        client.on('Runtime.executionContextsCleared', lambda event: self._onExecutionContextsCleared())
        client.on('Page.lifecycleEvent', lambda event: self._onLifecycleEvent(event))

    async def initialize(self) -> None:
        await self._client.send('Page.enable')
        frameTree = (await self._client.send('Page.getFrameTree'))['frameTree']
        self._handleFrameTree(frameTree)

        async def runtime_enabled() -> None:
            await self._client.send('Runtime.enable', {})
            await self._ensureIsolatedWorld(UTILITY_WORLD_NAME)

        await asyncio.gather(
            self._client.send('Page.setLifecycleEventsEnabled', {'enabled': True}),
            runtime_enabled(),
            self._networkManager.initialize(),
        )

    @property
    def networkManager(self) -> NetworkManager:
        return self._networkManager

    async def navigateFrame(
        self, frame: 'Frame', url: str, referer: str = None, timeout: float = None, waitUntil: WaitTargets = None,
    ) -> Optional[Response]:
        ensureNewDocumentNavigation = False

        async def navigate(url_: str, referer_: Optional[str], frameId: str) -> Optional[Exception]:
            try:
                # careful, this is the correct spelling of referrer (even though everywhere else the key is correct
                response = await self._client.send(
                    'Page.navigate', {'url': url_, 'referrer': referer_, 'frameId': frameId}
                )
                nonlocal ensureNewDocumentNavigation
                ensureNewDocumentNavigation = bool(response.get('loaderId'))
                if response.get('errorText'):
                    raise BrowserError(f'{response["errorText"]} at {url}')
            except Exception as e:
                return e

        if referer is None:
            referer = self._networkManager.extraHTTPHeaders().get('referer', '')
        if waitUntil is None:
            waitUntil = ['load']
        if timeout is None:
            timeout = self._timeoutSettings.navigationTimeout

        watcher = LifecycleWatcher(self, frame=frame, timeout=timeout, waitUntil=waitUntil)
        error = await helpers.future_race(navigate(url, referer, frame._id), watcher.timeoutOrTerminationFuture)
        if not error:
            if ensureNewDocumentNavigation:
                nav_fut = watcher.newDocumentNavigationFuture
            else:
                nav_fut = watcher.sameDocumentNavigationFuture
            error = await helpers.future_race(watcher.timeoutOrTerminationFuture, nav_fut)
        watcher.dispose()
        if error:
            raise error
        return watcher.navigationResponse()

    async def waitForFrameNavigation(
        self, frame: 'Frame', waitUntil: WaitTargets = None, timeout: float = None
    ) -> Optional[Response]:
        if not waitUntil:
            waitUntil = ['load']
        if not timeout:
            timeout = self._timeoutSettings.navigationTimeout
        watcher = LifecycleWatcher(self, frame=frame, timeout=timeout, waitUntil=waitUntil)
        error = await helpers.future_race(
            watcher.timeoutOrTerminationFuture,
            watcher.sameDocumentNavigationFuture,
            watcher.newDocumentNavigationFuture,
        )
        watcher.dispose()
        if error:
            raise error
        return watcher.navigationResponse()

    def _onLifecycleEvent(self, event: Dict) -> None:
        frame = self._frames.get(event['frameId'])
        if not frame:
            return
        frame._onLifecycleEvent(event['loaderId'], event['name'])
        self.emit(Events.FrameManager.LifecycleEvent, frame)

    def _onFrameStoppedLoading(self, frameId: str) -> None:
        frame = self._frames.get(frameId)
        if not frame:
            return
        frame._onLoadingStopped()
        self.emit(Events.FrameManager.LifecycleEvent, frame)

    def _handleFrameTree(self, frameTree: Dict) -> None:
        frame = frameTree['frame']
        if 'parentId' in frame:
            self._onFrameAttached(
                frame['id'], frame['parentId'],
            )
        self._onFrameNavigated(frame)
        for child in frameTree.get('childFrames', []):
            self._handleFrameTree(child)

    @property
    def page(self) -> 'Page':
        return self._page

    @property
    def mainFrame(self) -> Optional['Frame']:
        """Return main frame."""
        return self._mainFrame

    @property
    def frames(self) -> List['Frame']:
        """Return all frames."""
        return list(self._frames.values())

    def frame(self, frameId: Optional[str]) -> Optional['Frame']:
        """Return :class:`Frame` of ``frameId``."""
        return self._frames.get(frameId)

    def _onFrameAttached(self, frameId: str, parentFrameId: str) -> None:
        if frameId in self._frames:
            return
        parentFrame = self._frames.get(parentFrameId)
        frame = Frame(frameManager=self, client=self._client, parentFrame=parentFrame, frameId=frameId)
        self._frames[frameId] = frame
        self.emit(Events.FrameManager.FrameAttached, frame)

    def _onFrameNavigated(self, framePayload: dict) -> None:
        isMainFrame = not framePayload.get('parentId')
        if isMainFrame:
            frame = self._mainFrame
        else:
            frame = self._frames.get(framePayload.get('id', ''))
        if not (isMainFrame or frame):
            raise PageError('We either navigate top level or have old version of the navigated frame')

        # Detach all child frames first.
        if frame:
            for child in frame.childFrames:
                self._removeFramesRecursively(child)

        # Update or create main frame.
        _id = framePayload.get('id', '')
        if isMainFrame:
            if frame:
                # Update frame id to retain frame identity on cross-process navigation.
                self._frames.pop(frame._id, None)
                frame._id = _id
            else:
                # Initial main frame navigation.
                frame = Frame(frameManager=self, client=self._client, parentFrame=None, frameId=_id)
            self._frames[_id] = frame
            self._mainFrame = frame

        # Update frame payload.
        frame._navigated(framePayload)  # type: ignore
        self.emit(Events.FrameManager.FrameNavigated, frame)

    async def _ensureIsolatedWorld(self, name: str) -> None:
        if name in self._isolatedWorlds:
            return
        self._isolatedWorlds.add(name)
        await self._client.send(
            'Page.addScriptToEvaluateOnNewDocument',
            {'source': f'//# sourceURL={EVALUATION_SCRIPT_URL}', 'worldName': name,},
        )
        results = await asyncio.gather(
            *[
                self._client.send(
                    'Page.createIsolatedWorld', {'frameId': frame._id, 'grantUniversalAccess': True, 'worldName': name,}
                )
                for frame in self.frames
            ],
            return_exceptions=True,
        )
        for result in results:
            if isinstance(result, Exception):
                logger.exception(f'An exception occurred: {result}')

    def _onFrameNavigatedWithinDocument(self, frameId: str, url: str) -> None:
        frame = self._frames.get(frameId)
        if not frame:
            return
        frame._navigatedWithinDocument(url)
        self.emit(Events.FrameManager.FrameNavigatedWithinDocument, frame)
        self.emit(Events.FrameManager.FrameNavigated, frame)

    def _onFrameDetached(self, frameId: str) -> None:
        frame = self._frames.get(frameId)
        if frame:
            self._removeFramesRecursively(frame)

    def _onExecutionContextCreated(self, contextPayload: Dict) -> None:
        auxData = contextPayload.get('auxData')
        frameId = auxData.get('frameId')
        frame = self._frames.get(frameId)
        world = None
        if frame:
            if auxData and auxData['isDefault']:
                world = frame._mainWorld
            elif contextPayload.get('name') == UTILITY_WORLD_NAME and not frame._secondaryWorld._hasContext():
                # In case of multiple sessions to the same target, there's a race between
                # connections so we might end up creating multiple isolated worlds.
                # We can use either.
                world = frame._secondaryWorld
        if auxData and auxData.get('type') == 'isolated':
            self._isolatedWorlds.add(contextPayload['name'])

        context = ExecutionContext(self._client, contextPayload, world)
        if world:
            world._setContext(context)
        self._contextIdToContext[contextPayload['id']] = context

    def _onExecutionContextDestroyed(self, executionContextId: int) -> None:
        context = self._contextIdToContext.get(executionContextId)
        if not context:
            return
        del self._contextIdToContext[executionContextId]
        if context._world:
            context._world._setContext(None)

    def _onExecutionContextsCleared(self) -> None:
        for context in self._contextIdToContext.values():
            if context._world:
                context._world._setContext(None)
        self._contextIdToContext.clear()

    def executionContextById(self, contextId: int) -> 'ExecutionContext':
        """Get stored ``ExecutionContext`` by ``id``."""
        context = self._contextIdToContext.get(contextId)
        if not context:
            raise ElementHandleError(f'INTERNAL ERROR: missing context with id = {contextId}')
        return context

    def _removeFramesRecursively(self, frame: 'Frame') -> None:
        for child in frame.childFrames:
            self._removeFramesRecursively(child)
        frame._detach()
        self._frames.pop(frame._id, None)
        self.emit(Events.FrameManager.FrameDetached, frame)



import json
import logging
import os
import time
import base64
import subprocess
from typing import Dict
from flask import Flask, request, Response, jsonify
from flask_cors import CORS
from biolib.biolib_binary_format import SavedJob
from biolib.compute_node.webserver import webserver_utils
from biolib.compute_node.webserver import webserver_config
from biolib.compute_node.webserver.gunicorn_flask_application import GunicornFlaskApplication

# Disable warning about using the "global" statement for the rest of this file
# pylint: disable=global-statement

biolib_logger = logging.getLogger('biolib')

# Global constants
DEV_MODE = os.getenv('COMPUTE_NODE_ENV') == 'dev'
EIF_PATH = ''
BIOLIB_HOST = ''
JOB_ID_TO_COMPUTE_STATE_DICT: Dict = {}
UNASSIGNED_COMPUTE_PROCESSES = []

app = Flask(__name__)
CORS(app)


def shutdown_after_response():
    webserver_utils.deregister()
    biolib_logger.debug("Shutting down...")
    if not DEV_MODE:
        subprocess.run(['sudo', 'shutdown', 'now'], check=True)


@app.route('/hello/')
def hello():
    return 'Hello'


@app.route('/v1/job/', methods=['POST'])
def start_job():
    global JOB_ID_TO_COMPUTE_STATE_DICT, BIOLIB_HOST
    saved_job = json.loads(request.data.decode())

    if not webserver_utils.validate_saved_job(saved_job):
        return jsonify({'job': 'Invalid job'}), 400

    job_id = saved_job['job']['public_id']
    saved_job['BIOLIB_HOST'] = BIOLIB_HOST

    compute_state = webserver_utils.get_compute_state(UNASSIGNED_COMPUTE_PROCESSES)
    JOB_ID_TO_COMPUTE_STATE_DICT[job_id] = compute_state

    if EIF_PATH:
        job_remote_hosts = [host['hostname'] for host in saved_job['job']['app_version']['remote_hosts']]
        webserver_utils.start_vsock_proxies(remote_hosts=job_remote_hosts)
        saved_job['enclave_ecr_token'] = os.getenv('ENCLAVE_ECR_TOKEN', None)
        if not DEV_MODE:
            # Cancel the long general timer and replace with shorter shutdown timer
            subprocess.run(['sudo', 'shutdown', '-c'], check=True)
            subprocess.run(
                ['sudo', 'shutdown', f'+{webserver_config.COMPUTE_NODE_RUNNING_JOB_SHUTDOWN_TIME_MINUTES}'], check=True)

    saved_job_bbf_package = SavedJob().serialize(json.dumps(saved_job))
    send_package_to_compute_process(job_id, saved_job_bbf_package)

    if EIF_PATH:
        return Response(base64.b64encode(compute_state['attestation_document']), status=201)
    else:
        return '', 201


@app.route('/v1/job/<job_id>/start/', methods=['POST'])
def start_compute(job_id):
    module_input_package = request.data
    send_package_to_compute_process(job_id, module_input_package)
    return '', 201


@app.route('/v1/job/<job_id>/status/')
def status(job_id):
    # TODO Implement auth token
    global JOB_ID_TO_COMPUTE_STATE_DICT
    current_status = JOB_ID_TO_COMPUTE_STATE_DICT[job_id]['status'].copy()

    if current_status['status_updates']:
        JOB_ID_TO_COMPUTE_STATE_DICT[job_id]['status']['status_updates'] = []

    # Check if any error occured
    if 'error_code' in current_status:
        if EIF_PATH:
            error_response = app.response_class(response=json.dumps(current_status),
                                                status=201,
                                                mimetype='application/json')
            error_response.call_on_close(shutdown_after_response)
            return error_response
        else:
            # Remove failed job
            JOB_ID_TO_COMPUTE_STATE_DICT.pop(job_id)

    return jsonify(current_status)


@app.route('/v1/job/<job_id>/result/')
def result(job_id):
    global JOB_ID_TO_COMPUTE_STATE_DICT
    if JOB_ID_TO_COMPUTE_STATE_DICT[job_id]['result']:
        result_data = JOB_ID_TO_COMPUTE_STATE_DICT[job_id]['result']
        result_response = Response(result_data)

        if EIF_PATH:
            result_response.call_on_close(shutdown_after_response)
        else:
            # Remove finished job
            JOB_ID_TO_COMPUTE_STATE_DICT.pop(job_id)

        return result_response
    else:
        return '', 404


def send_package_to_compute_process(job_id, package_bytes):
    global JOB_ID_TO_COMPUTE_STATE_DICT
    message_queue = JOB_ID_TO_COMPUTE_STATE_DICT[job_id]['messages_to_send_queue']
    message_queue.put(package_bytes)


def start_webserver(port, host, specified_biolib_host, specified_eif_path=None):
    global EIF_PATH, BIOLIB_HOST, UNASSIGNED_COMPUTE_PROCESSES, JOB_ID_TO_COMPUTE_STATE_DICT
    BIOLIB_HOST = specified_biolib_host

    if specified_eif_path:
        EIF_PATH = specified_eif_path

    if EIF_PATH and not DEV_MODE:
        subprocess.run(['sudo', 'shutdown', f'+{webserver_config.COMPUTE_NODE_AUTO_SHUTDOWN_TIME_MINUTES}'], check=True)

    def worker_exit(server, worker):  # pylint: disable=unused-argument
        active_compute_states = list(JOB_ID_TO_COMPUTE_STATE_DICT.values()) + UNASSIGNED_COMPUTE_PROCESSES
        biolib_logger.info(f"Sending kill signal to {len(active_compute_states)} compute processes")
        if active_compute_states:
            for compute_state in active_compute_states:
                if compute_state['worker_thread']:
                    compute_state['worker_thread'].terminate()
            time.sleep(4)
        return

    def post_fork(server, worker):  # pylint: disable=unused-argument
        global UNASSIGNED_COMPUTE_PROCESSES, EIF_PATH
        biolib_logger.info("Started Webserver")
        webserver_utils.start_compute_process(UNASSIGNED_COMPUTE_PROCESSES, EIF_PATH)

    options = {
        'bind': f'{host}:{port}',
        'workers': 1,
        'post_fork': post_fork,
        'worker_exit': worker_exit,
        'timeout': webserver_config.GUNICORN_REQUEST_TIMEOUT,
        'graceful_timeout': 4
    }

    # Log to file when running on EC2
    if EIF_PATH:
        options.update({
            'errorlog': webserver_config.LOG_FILE_PATH,
            'capture_output': True
        })

    GunicornFlaskApplication(app, options).run()



import importlib
import sys
import os
import logging
import argparse
from biolib.app import BioLibApp, BioLib

IS_DEV = os.getenv('BIOLIB_DEV', '').upper() == "TRUE"

def configure_logging(default_log_level):
    # set log level
    env_log_level = os.getenv('BIOLIB_LOG')
    if env_log_level is None:
        BioLib.set_logging(default_log_level)
    else:
        env_log_level_upper = env_log_level.upper()
        if env_log_level_upper == 'TRACE':
            BioLib.set_logging(BioLib.TRACE_LOGGING)
        elif env_log_level_upper == 'DEBUG':
            BioLib.set_logging(logging.DEBUG)
        elif env_log_level_upper == 'INFO':
            BioLib.set_logging(logging.INFO)
        elif env_log_level_upper == 'WARNING' or env_log_level_upper == 'WARN':
            BioLib.set_logging(logging.WARNING)
        else:
            BioLib.set_logging(logging.ERROR)

def cli():
    configure_logging(default_log_level=logging.ERROR)

    # set auth
    BioLib.set_api_token(os.getenv('BIOLIB_TOKEN'))

    # set host
    BioLib.set_host(os.getenv('BIOLIB_HOST'))

    if len(sys.argv) > 2 and sys.argv[1] == 'run':
        app_splitted = sys.argv[2].split('/')
        if len(app_splitted) == 2:
            app_author = app_splitted[0]
            app_name = app_splitted[1]
            app = BioLibApp(author=app_author, name=app_name)
            stdin = None
            if not sys.stdin.isatty() and not IS_DEV:
                stdin = sys.stdin.read()
            app_args = sys.argv[3:]
            result = app(args=app_args, stdin=stdin, files=None)
            sys.stdout.buffer.write(result.stdout)
            sys.stderr.buffer.write(result.stderr)
        else:
            print(f'App name {sys.argv[2]} was incorrectly formatted. Please use this format: app_developer/app_name')
    else:

        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers(help='command', dest='command')

        # Add subparser for run to help message makes sense
        # The actual code for running applications is above this
        _parser_run = subparsers.add_parser('run', help='Run an application on BioLib')

        # add subparser for push
        parser_push = subparsers.add_parser('push', help='Push an application to BioLib')
        parser_push.add_argument('author_and_app_name')
        parser_push.add_argument('--path', default='.', required=False)

        # add subparser for build-enclave
        parser_build_enclave = subparsers.add_parser('build-enclave', help='Build a nitro enclave EIF file')
        parser_build_enclave.add_argument('--dev', required=False, action='store_true')

        parser_compute_process = subparsers.add_parser('start-compute-process', help='Start a compute process')
        parser_compute_process.add_argument('--enclave', required=False, action='store_true')
        parser_compute_process.add_argument('--port', required=False, type=port_number)

        # add subparser for run-compute-node
        parser_start = subparsers.add_parser('start', help='Start a compute node')
        parser_start.add_argument('--eif', required=False, type=file_path)
        parser_start.add_argument('--port', default='5000', required=False, type=port_number)
        parser_start.add_argument('--host', default='127.0.0.1', required=False)

        args = parser.parse_args()

        if args.command == 'push':
            # always use INFO logging for push
            configure_logging(default_log_level=logging.INFO)
            BioLib.push(args.author_and_app_name, args.path)
        elif args.command == 'build-enclave':
            BioLib.build_enclave(args.dev)
        elif args.command == 'start':
            BioLib.start_compute_node(args.port, args.host, args.eif)
        elif args.command == 'start-compute-process':
            BioLib.start_compute_process(socket_port=args.port, is_running_in_enclave=args.enclave)
        else:
            print('Unrecognized command, please run biolib --help to see available options.')
            sys.exit(1)


class IllegalArgumentError(ValueError):
    pass


def file_path(path):
    if path.startswith('/'):
        full_path = path
    else:
        full_path = os.path.normpath(os.path.join(os.getcwd(), path))
    if not os.path.exists(full_path):
        raise IllegalArgumentError(f'The path {full_path} does not exist')
    return full_path


def port_number(port):
    if not port.isdigit():
        raise IllegalArgumentError(f'Port number {port} is not a number. Ports can only be numbers')

    if not (0 < int(port) < 65000):  # pylint: disable=superfluous-parens
        raise IllegalArgumentError('Port can only be between 0 and 65000')

    return port


class ImportHook(object):

    # The argument 'path' is not used, but it must be defined since importlib expects this function interface
    def find_module(self, fullname, path=None):  # pylint: disable=unused-argument
        # import hook for all imports in the form blb.*
        if fullname.split('.')[0] == 'biolib':
            return self
        else:
            return None

    def load_module(self, fullname):
        fullname_splitted = fullname.split('.')
        assert fullname_splitted[0] == 'biolib'

        # don't override existing module
        if fullname in sys.modules:
            return sys.modules[fullname]

        # dynamically create new module
        if len(fullname_splitted) == 3:
            sys.modules[fullname] = BioLibApp(author=fullname_splitted[1], name=fullname_splitted[2])
            return sys.modules[fullname]
        elif len(fullname_splitted) == 2:
            spec = importlib.util.spec_from_file_location(fullname, __file__)
            mod = importlib.util.module_from_spec(spec)
            sys.modules[fullname] = mod
        else:
            raise Exception(f'Import `{fullname}` incorrectly formatted')
        return mod


sys.meta_path = [ImportHook()]

if IS_DEV and __name__ == '__main__':
    cli()



from biolib.biolib_binary_format.base_bbf_package import BioLibBinaryFormatBasePackage


class ModuleInput(BioLibBinaryFormatBasePackage):
    def __init__(self, bbf=None):
        super().__init__(bbf)
        self.package_type = 1

    def serialize(self, stdin, arguments, files):
        bbf_data = bytearray()
        bbf_data.extend(self.version.to_bytes(1, 'big'))
        bbf_data.extend(self.package_type.to_bytes(1, 'big'))

        bbf_data.extend(len(stdin).to_bytes(8, 'big'))

        argument_len = sum([len(arg.encode()) for arg in arguments]) + (2 * len(arguments))
        bbf_data.extend(argument_len.to_bytes(4, 'big'))

        file_data_len = sum([len(data) + len(path.encode()) for path, data in files.items()]) + (12 * len(files))
        bbf_data.extend(file_data_len.to_bytes(8, 'big'))

        bbf_data.extend(stdin)

        for argument in arguments:
            bbf_data.extend(len(argument).to_bytes(2, 'big'))
            bbf_data.extend(argument.encode())

        for path, data in files.items():
            bbf_data.extend(len(path.encode()).to_bytes(4, 'big'))
            bbf_data.extend(len(data).to_bytes(8, 'big'))

            bbf_data.extend(path.encode())
            bbf_data.extend(data)

        return bbf_data

    def deserialize(self):
        version = self.get_data(1, output_type='int')
        package_type = self.get_data(1, output_type='int')
        self.check_version_and_type(version=version, package_type=package_type, expected_package_type=self.package_type)

        stdin_len = self.get_data(8, output_type='int')
        argument_data_len = self.get_data(4, output_type='int')
        files_data_len = self.get_data(8, output_type='int')
        stdin = self.get_data(stdin_len)

        end_of_arguments = self.pointer + argument_data_len
        arguments = []
        while self.pointer != end_of_arguments:
            argument_len = self.get_data(2, output_type='int')
            argument = self.get_data(argument_len, output_type='str')
            arguments.append(argument)

        end_of_files = self.pointer + files_data_len
        files = {}
        while self.pointer < end_of_files:
            path_len = self.get_data(4, output_type='int')
            data_len = self.get_data(8, output_type='int')
            path = self.get_data(path_len, output_type='str')
            data = self.get_data(data_len)
            files[path] = bytes(data)

        return {'stdin': stdin, 'arguments': arguments, 'files': files}



import socket
import sys
import threading
import errno
from typing import Optional
from queue import Queue


class SocketListenerThread(threading.Thread):
    DEFAULT_BUFFER_SIZE = 1024

    def __init__(self, socket_connection: socket.socket, message_queue: Queue):
        threading.Thread.__init__(self)

        self.socket_connection = socket_connection
        self.message_queue = message_queue

    def run(self) -> None:
        while True:
            try:
                next_package_length_as_bytearray = self._receive_specific_byte_length(8)
            except IOError as exception:
                # [Errno 9] Bad file descriptor
                if exception.errno == errno.EBADF:
                    break
                else:
                    raise exception

            if next_package_length_as_bytearray is not None:
                next_package_length_as_int = int.from_bytes(next_package_length_as_bytearray, 'big')
                package = self._receive_specific_byte_length(next_package_length_as_int)
                self.message_queue.put(package)

    def _receive_specific_byte_length(self, byte_length_to_receive: int) -> Optional[bytearray]:
        return_data = bytearray()

        while len(return_data) < byte_length_to_receive:
            byte_length_left_to_receive = byte_length_to_receive - len(return_data)
            packet = self.socket_connection.recv(min(self.DEFAULT_BUFFER_SIZE, byte_length_left_to_receive))
            if not packet:
                return None
            return_data.extend(packet)

        return return_data

    def terminate(self):
        sys.exit()
